#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import warnings
import time

import numpy as np
import pandas as pd

from mars.utils import parse_readable_size, ceildiv
from mars.serialization.serializables import (
    StringField,
    Int64Field,
    ListField,
    SeriesField,
    DictField,
    BoolField,
    AnyField,
    FieldTypes,
)
from mars.dataframe import ArrowStringDtype
from mars.dataframe.core import DATAFRAME_CHUNK_TYPE
from mars.dataframe.utils import (
    parse_index,
    standardize_range_index,
    arrow_table_to_pandas_dataframe,
)
from mars.dataframe.datasource.core import ColumnPruneSupportedDataSourceMixin
from mars.core import OutputType

from ....config import options
from ....df.backends.odpssql.types import odps_type_to_df_type
from ....utils import to_str, to_timestamp
from ...utils import check_partition_exist, filter_partitions
from ..cupid_service import CupidServiceClient

logger = logging.getLogger(__name__)

CHUNK_BYTES_LIMIT = 64 * 1024**2
MAX_CHUNK_NUM = 512 * 1024**2

ORC_COMPRESSION_RATIO = 5
STRING_FIELD_OVERHEAD = 50

try:
    from mars.dataframe.datasource.core import (
        IncrementalIndexDatasource,
        IncrementalIndexDataSourceMixin,
    )

    _BASE = (
        IncrementalIndexDatasource,
        ColumnPruneSupportedDataSourceMixin,
        IncrementalIndexDataSourceMixin,
    )
    _NEED_STANDARDIZE = False
except ImportError:
    from mars.dataframe.datasource.core import HeadOptimizedDataSource

    _BASE = (HeadOptimizedDataSource, ColumnPruneSupportedDataSourceMixin)
    _NEED_STANDARDIZE = True


class DataFrameReadTable(*_BASE):
    _op_type_ = 123450

    odps_params = DictField("odps_params")
    table_name = StringField("table_name")
    partition_spec = StringField("partition_spec", default=None)
    dtypes = SeriesField("dtypes", default=None)
    index_type = StringField("index_type", default=None)
    columns = AnyField("columns", default=None)
    nrows = Int64Field("nrows", default=None)
    use_arrow_dtype = BoolField("use_arrow_dtype", default=None)
    string_as_binary = BoolField("string_as_binary", default=None)
    append_partitions = BoolField("append_partitions", default=None)
    last_modified_time = Int64Field("last_modified_time", default=None)
    with_split_meta_on_tile = BoolField("with_split_meta_on_tile", default=None)
    retry_times = Int64Field("retry_times", default=None)
    tunnel_quota_name = StringField("tunnel_quota_name", default=None)
    index_columns = ListField("index_columns", FieldTypes.string, default=None)
    index_dtypes = SeriesField("index_dtypes", default=None)

    def __init__(self, sparse=None, memory_scale=None, **kw):
        super(DataFrameReadTable, self).__init__(
            sparse=sparse,
            memory_scale=memory_scale,
            _output_types=[OutputType.dataframe],
            **kw
        )

    @property
    def retryable(self):
        if "CUPID_SERVICE_SOCKET" in os.environ:
            return False
        else:
            return True

    @property
    def partition(self):
        return getattr(self, "partition_spec", None)

    @property
    def incremental_index(self):
        return self.index_type == "incremental"

    def get_columns(self):
        return self.columns

    def set_pruned_columns(self, columns, *, keep_order=None):  # pragma: no cover
        self.columns = columns

    def __call__(self, shape, chunk_bytes=None, chunk_size=None):
        import numpy as np
        import pandas as pd

        if not self.index_columns:
            if np.isnan(shape[0]):
                index_value = parse_index(pd.RangeIndex(0))
            else:
                index_value = parse_index(pd.RangeIndex(shape[0]))
        elif len(self.index_columns) == 1:
            index_value = parse_index(pd.Index([]).astype(self.index_dtypes[0]))
        else:
            idx = pd.MultiIndex.from_frame(
                pd.DataFrame([], columns=self.index_columns).astype(self.index_dtypes)
            )
            index_value = parse_index(idx)

        columns_value = parse_index(self.dtypes.index, store_data=True)
        return self.new_dataframe(
            None,
            shape,
            dtypes=self.dtypes,
            index_value=index_value,
            columns_value=columns_value,
            chunk_bytes=chunk_bytes,
            chunk_size=chunk_size,
        )

    @classmethod
    def _tile_cupid(cls, op):
        import numpy as np
        import pandas as pd
        from mars.core.context import get_context

        df = op.outputs[0]
        split_size = df.extra_params.chunk_bytes or CHUNK_BYTES_LIMIT

        out_dtypes = df.dtypes
        out_shape = df.shape
        out_columns_value = df.columns_value
        if op.columns is not None:
            out_dtypes = out_dtypes[op.columns]
            out_shape = (df.shape[0], len(op.columns))
            out_columns_value = parse_index(out_dtypes.index, store_data=True)

        mars_context = get_context()
        if mars_context is not None:
            worker_count = len(mars_context.get_worker_addresses())
        else:
            worker_count = None

        cupid_client = CupidServiceClient()
        try:
            parts = cupid_client.enum_table_partitions(
                op.odps_params, op.table_name, op.partition
            )
            if parts is None:
                parts = [None]

            out_chunks = []
            chunk_idx = 0

            for partition_spec in parts:
                columns = op.columns
                if columns is not None:
                    columns = (op.index_columns or []) + columns
                splits, split_size = cupid_client.create_table_download_session(
                    op.odps_params,
                    op.table_name,
                    partition_spec,
                    columns,
                    worker_count,
                    split_size,
                    MAX_CHUNK_NUM,
                    op.with_split_meta_on_tile,
                )

                logger.info("%s table splits have been created.", str(len(splits)))
                meta_chunk_rows = [split.meta_row_count for split in splits]
                if np.isnan(out_shape[0]):
                    est_chunk_rows = meta_chunk_rows
                else:
                    sp_file_sizes = np.array(
                        [sp.split_file_end - sp.split_file_start for sp in splits]
                    )
                    total_size = sp_file_sizes.sum()
                    ratio_chunk_rows = (
                        sp_file_sizes * out_shape[0] // total_size
                    ).tolist()
                    est_chunk_rows = [
                        mr if mr is not None else rr
                        for mr, rr in zip(meta_chunk_rows, ratio_chunk_rows)
                    ]

                logger.warning("Estimated chunk rows: %r", est_chunk_rows)

                if len(splits) == 0:
                    logger.info("Table %s has no data", op.table_name)
                    chunk_op = DataFrameReadTableSplit()
                    index_value = parse_index(pd.RangeIndex(0))
                    columns_value = parse_index(out_dtypes.index, store_data=True)
                    out_chunk = chunk_op.new_chunk(
                        None,
                        shape=(np.nan, out_shape[1]),
                        dtypes=op.dtypes,
                        index_value=index_value,
                        columns_value=columns_value,
                        index=(chunk_idx, 0),
                    )
                    out_chunks.append(out_chunk)
                    chunk_idx += 1
                else:
                    for idx, split in enumerate(splits):
                        chunk_op = DataFrameReadTableSplit(
                            cupid_handle=to_str(split.handle),
                            split_index=split.split_index,
                            split_file_start=split.split_file_start,
                            split_file_end=split.split_file_end,
                            schema_file_start=split.schema_file_start,
                            schema_file_end=split.schema_file_end,
                            index_type=op.index_type,
                            dtypes=out_dtypes,
                            sparse=op.sparse,
                            split_size=split_size,
                            string_as_binary=op.string_as_binary,
                            use_arrow_dtype=op.use_arrow_dtype,
                            estimate_rows=est_chunk_rows[idx],
                            partition_spec=partition_spec,
                            append_partitions=op.append_partitions,
                            meta_raw_size=split.meta_raw_size,
                            nrows=meta_chunk_rows[idx] or op.nrows,
                            memory_scale=op.memory_scale,
                            index_columns=op.index_columns,
                            extra_params=op.extra_params,
                        )
                        # the chunk shape is unknown
                        index_value = parse_index(pd.RangeIndex(0))
                        columns_value = parse_index(out_dtypes.index, store_data=True)
                        out_chunk = chunk_op.new_chunk(
                            None,
                            shape=(np.nan, out_shape[1]),
                            dtypes=out_dtypes,
                            index_value=index_value,
                            columns_value=columns_value,
                            index=(chunk_idx, 0),
                        )
                        chunk_idx += 1
                        out_chunks.append(out_chunk)
        finally:
            cupid_client.close()

        if op.index_type == "incremental" and _NEED_STANDARDIZE:
            out_chunks = standardize_range_index(out_chunks)
        new_op = op.copy()
        nsplits = ((np.nan,) * len(out_chunks), (out_shape[1],))
        return new_op.new_dataframes(
            None,
            shape=out_shape,
            dtypes=op.dtypes,
            index_value=df.index_value,
            columns_value=out_columns_value,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    @classmethod
    def _tile_tunnel(cls, op):
        from odps import ODPS

        project = os.environ.get("ODPS_PROJECT_NAME", None)
        odps_params = op.odps_params.copy()
        if project:
            odps_params["project"] = project
        endpoint = os.environ.get("ODPS_RUNTIME_ENDPOINT") or odps_params["endpoint"]
        o = ODPS(
            odps_params["access_id"],
            odps_params["secret_access_key"],
            project=odps_params["project"],
            endpoint=endpoint,
        )

        table_obj = o.get_table(op.table_name)
        if not table_obj.table_schema.partitions:
            data_srcs = [table_obj]
        elif op.partition is not None and check_partition_exist(
            table_obj, op.partition
        ):
            data_srcs = [table_obj.get_partition(op.partition)]
        else:
            data_srcs = list(table_obj.partitions)
            if op.partition is not None:
                data_srcs = filter_partitions(o, data_srcs, op.partition)

        out_chunks = []
        row_nsplits = []
        index_start = 0
        df = op.outputs[0]

        out_dtypes = df.dtypes
        out_shape = df.shape
        out_columns_value = df.columns_value
        if op.columns is not None:
            out_dtypes = out_dtypes[op.columns]
            out_shape = (df.shape[0], len(op.columns))
            out_columns_value = parse_index(out_dtypes.index, store_data=True)

        if len(data_srcs) == 0:
            # no partitions are selected
            chunk_op = DataFrameReadTableSplit()
            index_value = parse_index(pd.RangeIndex(0))
            columns_value = parse_index(out_dtypes.index, store_data=True)
            out_chunk = chunk_op.new_chunk(
                None,
                shape=(0, out_shape[1]),
                dtypes=op.dtypes,
                index_value=index_value,
                columns_value=columns_value,
                index=(index_start, 0),
            )
            out_chunks.append(out_chunk)
        else:
            retry_times = op.retry_times or options.retry_times
            for data_src in data_srcs:
                data_store_size = data_src.size

                retries = 0
                while True:
                    try:
                        with data_src.open_reader() as reader:
                            record_count = reader.count
                        break
                    except:
                        if retries >= retry_times:
                            raise
                        retries += 1
                        time.sleep(1)
                if data_store_size == 0:
                    # empty table
                    chunk_op = DataFrameReadTableSplit()
                    index_value = parse_index(pd.RangeIndex(0))
                    columns_value = parse_index(out_dtypes.index, store_data=True)
                    out_chunk = chunk_op.new_chunk(
                        None,
                        shape=(0, out_shape[1]),
                        dtypes=op.dtypes,
                        index_value=index_value,
                        columns_value=columns_value,
                        index=(index_start, 0),
                    )
                    out_chunks.append(out_chunk)
                    index_start += 1
                    continue
                chunk_size = df.extra_params.chunk_size

                partition_spec = (
                    str(data_src.partition_spec)
                    if getattr(data_src, "partition_spec", None)
                    else None
                )

                if chunk_size is None:
                    chunk_bytes = df.extra_params.chunk_bytes or CHUNK_BYTES_LIMIT
                    chunk_count = data_store_size // chunk_bytes + (
                        data_store_size % chunk_bytes != 0
                    )
                    chunk_size = ceildiv(record_count, chunk_count)
                    split_size = chunk_bytes
                else:
                    chunk_count = ceildiv(record_count, chunk_size)
                    split_size = data_store_size // chunk_count

                for i in range(chunk_count):
                    start_index = chunk_size * i
                    end_index = min(chunk_size * (i + 1), record_count)
                    row_size = end_index - start_index
                    chunk_op = DataFrameReadTableSplit(
                        table_name=op.table_name,
                        partition_spec=partition_spec,
                        start_index=start_index,
                        end_index=end_index,
                        nrows=op.nrows,
                        odps_params=op.odps_params,
                        columns=op.columns,
                        index_type=op.index_type,
                        dtypes=out_dtypes,
                        sparse=op.sparse,
                        split_size=split_size,
                        use_arrow_dtype=op.use_arrow_dtype,
                        estimate_rows=row_size,
                        append_partitions=op.append_partitions,
                        memory_scale=op.memory_scale,
                        retry_times=op.retry_times,
                        tunnel_quota_name=op.tunnel_quota_name,
                        index_columns=op.index_columns,
                        extra_params=op.extra_params,
                    )
                    index_value = parse_index(pd.RangeIndex(start_index, end_index))
                    columns_value = parse_index(out_dtypes.index, store_data=True)
                    out_chunk = chunk_op.new_chunk(
                        None,
                        shape=(row_size, out_shape[1]),
                        dtypes=out_dtypes,
                        index_value=index_value,
                        columns_value=columns_value,
                        index=(index_start + i, 0),
                    )
                    row_nsplits.append(row_size)
                    out_chunks.append(out_chunk)

                index_start += chunk_count

        if op.index_type == "incremental" and _NEED_STANDARDIZE:
            out_chunks = standardize_range_index(out_chunks)

        new_op = op.copy()
        nsplits = (tuple(row_nsplits), (out_shape[1],))
        return new_op.new_dataframes(
            None,
            shape=out_shape,
            dtypes=op.dtypes,
            index_value=df.index_value,
            columns_value=out_columns_value,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    @classmethod
    def _tile(cls, op):
        if "CUPID_SERVICE_SOCKET" in os.environ:
            return cls._tile_cupid(op)
        else:
            return cls._tile_tunnel(op)


class DataFrameReadTableSplit(*_BASE):
    _op_type_ = 123451

    # for cupid
    cupid_handle = StringField("cupid_handle", default=None)
    split_index = Int64Field("split_index", default=None)
    split_file_start = Int64Field("split_file_start", default=None)
    split_file_end = Int64Field("split_file_end", default=None)
    schema_file_start = Int64Field("schema_file_start", default=None)
    schema_file_end = Int64Field("schema_file_end", default=None)
    use_arrow_dtype = BoolField("use_arrow_dtype", default=None)
    string_as_binary = BoolField("string_as_binary", default=None)
    dtypes = SeriesField("dtypes", default=None)
    nrows = Int64Field("nrows", default=None)
    index_type = StringField("index_type", default=None)

    # for tunnel
    table_name = StringField("table_name", default=None)
    partition_spec = StringField("partition_spec", default=None)
    start_index = Int64Field("start_index", default=None)
    end_index = Int64Field("end_index", default=None)
    odps_params = DictField("odps_params", default=None)
    columns = AnyField("columns", default=None)
    tunnel_quota_name = StringField("tunnel_quota_name", default=None)

    split_size = Int64Field("split_size", default=None)
    append_partitions = BoolField("append_partitions", default=None)
    estimate_rows = Int64Field("estimate_rows", default=None)
    meta_raw_size = Int64Field("meta_raw_size", default=None)
    retry_times = Int64Field("retry_times", default=None)
    index_columns = ListField("index_columns", FieldTypes.string, default=None)

    def __init__(self, memory_scale=None, sparse=None, **kw):
        super(DataFrameReadTableSplit, self).__init__(
            sparse=sparse,
            memory_scale=memory_scale,
            _output_types=[OutputType.dataframe],
            **kw
        )

    @property
    def retryable(self):
        if "CUPID_SERVICE_SOCKET" in os.environ:
            return False
        else:
            return True

    @property
    def output_limit(self):
        return 1

    @property
    def incremental_index(self):
        return self.index_type == "incremental"

    def set_pruned_columns(self, columns, *, keep_order=None):  # pragma: no cover
        if isinstance(columns, str):
            columns = [columns]
        self.columns = list(columns)

    def get_columns(self):
        return self.columns

    @classmethod
    def estimate_size(cls, ctx, op):
        import numpy as np

        def is_object_dtype(dtype):
            try:
                return (
                    np.issubdtype(dtype, np.object_)
                    or np.issubdtype(dtype, np.unicode_)
                    or np.issubdtype(dtype, np.bytes_)
                )
            except TypeError:  # pragma: no cover
                return False

        if op.split_size is None:
            ctx[op.outputs[0].key] = (0, 0)
            return

        arrow_size = (op.memory_scale or ORC_COMPRESSION_RATIO) * op.split_size
        if op.meta_raw_size is not None:
            raw_arrow_size = (op.memory_scale or 1) * op.meta_raw_size
            arrow_size = max(arrow_size, raw_arrow_size)

        n_strings = len([dt for dt in op.dtypes if is_object_dtype(dt)])
        if op.estimate_rows or op.nrows:
            rows = op.nrows if op.nrows is not None else op.estimate_rows
            pd_size = arrow_size + n_strings * rows * STRING_FIELD_OVERHEAD
            logger.info("Estimate pandas memory cost: %r", pd_size)
        else:
            pd_size = arrow_size * 10 if n_strings else arrow_size

        ctx[op.outputs[0].key] = (pd_size, pd_size + arrow_size)

    @classmethod
    def _cast_string_to_binary(cls, arrow_table):
        import pyarrow as pa

        new_schema = []
        for field in arrow_table.schema:
            if field.type == pa.string():
                new_schema.append(pa.field(field.name, pa.binary()))
            else:
                new_schema.append(field)

        return arrow_table.cast(pa.schema(new_schema))

    @classmethod
    def _append_partition_values(cls, arrow_table, op):
        import pyarrow as pa

        if op.append_partitions and op.partition_spec:
            from odps.types import PartitionSpec

            spec = PartitionSpec(op.partition_spec)

            for col_name, pt_val in spec.items():
                arrow_table = arrow_table.append_column(
                    col_name, pa.array([pt_val] * arrow_table.num_rows, pa.string())
                )

        return arrow_table

    @staticmethod
    def _align_columns(data, expected_dtypes):
        data_columns = data.dtypes.index
        expected_columns = expected_dtypes.index
        if not data_columns.equals(expected_columns):
            logger.info(
                "Data columns differs from output columns, "
                "data columns: %s, output columns: %s",
                data_columns,
                expected_columns,
            )
            data.columns = expected_columns[: len(data.columns)]
            for extra_col in expected_columns[len(data.columns) :]:
                data[extra_col] = pd.Series([], dtype=expected_dtypes[extra_col])
            if not data.dtypes.index.equals(expected_columns):
                data = data[expected_columns]
        return data

    @classmethod
    def _align_output_data(cls, op, data):
        if isinstance(op.outputs[0], DATAFRAME_CHUNK_TYPE):
            dtypes = op.outputs[0].dtypes
            data = cls._align_columns(data, dtypes)
        else:
            dtypes = pd.Series([op.outputs[0].dtype], index=[op.outputs[0].name])
            data = cls._align_columns(data, dtypes)
            data = data[op.outputs[0].name]
        return data

    @classmethod
    def _build_empty_df(cls, out):
        empty_df = pd.DataFrame()
        for name, dtype in out.dtypes.items():
            empty_df[name] = pd.Series(dtype=dtype)
        return empty_df

    @classmethod
    def _execute_in_cupid(cls, ctx, op):
        out = op.outputs[0]

        if op.cupid_handle is None:
            ctx[out.key] = cls._build_empty_df(out)
            return

        split_config = dict(
            _handle=op.cupid_handle,
            _split_index=op.split_index,
            _split_file_start=op.split_file_start,
            _split_file_end=op.split_file_end,
            _schema_file_start=op.schema_file_start,
            _schema_file_end=op.schema_file_end,
        )
        cupid_client = CupidServiceClient()
        try:
            pa_table = cupid_client.read_table_data(split_config, op.nrows)
        finally:
            cupid_client.close()
            cupid_client = None
        pa_table = cls._append_partition_values(pa_table, op)

        if op.string_as_binary:
            pa_table = cls._cast_string_to_binary(pa_table)
        data = arrow_table_to_pandas_dataframe(
            pa_table, use_arrow_dtype=op.use_arrow_dtype
        )[: op.nrows]
        if op.index_columns:
            data = data.set_index(op.index_columns)

        data = cls._align_output_data(op, data)
        logger.info("Read split table finished, split index: %s", op.split_index)
        logger.info("Split data shape is %s", data.shape)
        ctx[out.key] = data

    @classmethod
    def _execute_arrow_tunnel(cls, ctx, op):
        from odps import ODPS
        from odps.tunnel import TableTunnel

        out = op.outputs[0]

        if op.table_name is None:
            # is empty table
            ctx[out.key] = cls._build_empty_df(out)
            return

        project = os.environ.get("ODPS_PROJECT_NAME", None)
        odps_params = op.odps_params.copy()
        if project:
            odps_params["project"] = project
        endpoint = os.environ.get("ODPS_RUNTIME_ENDPOINT") or odps_params["endpoint"]
        o = ODPS(
            odps_params["access_id"],
            odps_params["secret_access_key"],
            project=odps_params["project"],
            endpoint=endpoint,
        )

        t = o.get_table(op.table_name)
        schema_name = t.get_schema().name if t.get_schema() else None
        tunnel = TableTunnel(o, project=t.project, quota_name=op.tunnel_quota_name)
        retry_times = op.retry_times or options.retry_times
        init_sleep_secs = 1
        logger.info(
            "Start creating download session for table %s(%s) start index %s end index %s retry_times %s.",
            op.table_name,
            op.partition_spec,
            op.start_index,
            op.end_index,
            retry_times,
        )
        retries = 0
        while True:
            try:
                if op.partition_spec is not None:
                    download_session = tunnel.create_download_session(
                        t.name,
                        partition_spec=op.partition_spec,
                        schema=schema_name,
                        async_mode=True,
                    )
                else:
                    download_session = tunnel.create_download_session(
                        t.name, async_mode=True
                    )
                break
            except:
                if retries >= retry_times:
                    raise
                retries += 1
                sleep_secs = retries * init_sleep_secs
                logger.exception(
                    "Create download session failed, sleep %s seconds and retry it",
                    sleep_secs,
                    exc_info=1,
                )
                time.sleep(sleep_secs)

        logger.info(
            "Start reading table %s(%s) split from %s to %s",
            op.table_name,
            op.partition_spec,
            op.start_index,
            op.end_index,
        )
        if op.nrows is None:
            count = op.end_index - op.start_index
        else:
            count = op.nrows

        retries = 0
        while True:
            try:
                columns = op.columns
                if columns is not None:
                    columns = (op.index_columns or []) + op.columns
                with download_session.open_arrow_reader(
                    op.start_index, count, columns=columns
                ) as reader:
                    table = reader.read()
                break
            except:
                if retries >= retry_times:
                    raise
                retries += 1
                sleep_secs = retries * init_sleep_secs
                logger.exception(
                    "Read table failed, sleep %s seconds and retry it",
                    sleep_secs,
                    exc_info=1,
                )
                time.sleep(sleep_secs)

        table = cls._append_partition_values(table, op)
        if op.string_as_binary:
            table = cls._cast_string_to_binary(table)
        data = arrow_table_to_pandas_dataframe(
            table, use_arrow_dtype=op.use_arrow_dtype
        )
        if op.index_columns:
            data = data.set_index(op.index_columns)
        data = cls._align_output_data(op, data)
        logger.info(
            "Finish reading table %s(%s) split from %s to %s",
            op.table_name,
            op.partition_spec,
            op.start_index,
            op.end_index,
        )
        ctx[op.outputs[0].key] = data

    @classmethod
    def execute(cls, ctx, op):
        if "CUPID_SERVICE_SOCKET" in os.environ:
            cls._execute_in_cupid(ctx, op)
        else:
            cls._execute_arrow_tunnel(ctx, op)


def df_type_to_np_type(df_type, use_arrow_dtype=False):
    from ....df import types
    from ....df.backends.pd.types import _df_to_np_types

    if df_type == types.string:
        if use_arrow_dtype:
            return ArrowStringDtype()
        else:
            return np.dtype("object")
    elif df_type in _df_to_np_types:
        return _df_to_np_types[df_type]
    elif df_type == types.timestamp:
        return np.datetime64(0, "ns").dtype
    else:
        return np.dtype("object")


def read_odps_table(
    table,
    shape,
    partition=None,
    sparse=False,
    chunk_bytes=None,
    chunk_size=None,
    columns=None,
    odps_params=None,
    add_offset=None,
    use_arrow_dtype=False,
    string_as_binary=None,
    memory_scale=None,
    append_partitions=False,
    with_split_meta_on_tile=False,
    index_type="incremental",
    tunnel_quota_name=None,
    index_columns=None,
    extra_params=None,
):
    import pandas as pd

    if add_offset is not None:
        warnings.warn(
            "add_offset is deprecated, please use index_type instead",
            DeprecationWarning,
        )
        if add_offset in (True, False):
            index_type = "incremental" if add_offset else "chunk_incremental"

    if isinstance(chunk_size, (list, tuple)):
        if len(chunk_size) == 1:
            chunk_size = chunk_size[0]
        if len(chunk_size) > 1:
            raise ValueError("Only support split on rows")

    if chunk_bytes is not None:
        chunk_bytes = int(parse_readable_size(chunk_bytes)[0])
    cols = table.table_schema.columns if append_partitions else table.table_schema.simple_columns
    table_columns = [c.name for c in cols]
    table_types = [c.type for c in cols]
    df_types = [
        df_type_to_np_type(odps_type_to_df_type(type), use_arrow_dtype=use_arrow_dtype)
        for type in table_types
    ]

    if isinstance(index_columns, str):
        index_columns = [index_columns]
    if index_columns and columns is None:
        index_col_set = set(index_columns)
        columns = [c for c in table_columns if c not in index_col_set]

    if not index_columns:
        index_dtypes = None
    else:
        table_index_types = [
            df_types[table_columns.index(col)] for col in index_columns
        ]
        index_dtypes = pd.Series(table_index_types, index=index_columns)

    if columns is not None:
        table_col_set = set(columns)
        if any(col in table_col_set for col in index_columns):
            raise ValueError("Index columns and columns shall not overlap.")

        # reorder columns
        new_columns = [c for c in table_columns if c in columns]
        df_types = [df_types[table_columns.index(col)] for col in new_columns]
        table_columns = new_columns
        columns = new_columns

    dtypes = pd.Series(df_types, index=table_columns)
    retry_times = options.retry_times
    op = DataFrameReadTable(
        odps_params=odps_params,
        table_name=table.full_table_name,
        partition_spec=partition,
        dtypes=dtypes,
        sparse=sparse,
        index_type=index_type,
        columns=columns,
        use_arrow_dtype=use_arrow_dtype,
        string_as_binary=string_as_binary,
        memory_scale=memory_scale,
        append_partitions=append_partitions,
        last_modified_time=to_timestamp(table.last_data_modified_time),
        with_split_meta_on_tile=with_split_meta_on_tile,
        retry_times=retry_times,
        tunnel_quota_name=tunnel_quota_name,
        index_columns=index_columns,
        index_dtypes=index_dtypes,
        extra_params=extra_params or dict(),
    )
    return op(shape, chunk_bytes=chunk_bytes, chunk_size=chunk_size)



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A couple of authentication types in ODPS.
"""

import base64
import hmac
import hashlib
import logging
import os
import threading
import time
from collections import OrderedDict
from datetime import datetime, timedelta

import requests

from .compat import six, cgi, urlparse, unquote, parse_qsl
from . import options, utils


logger = logging.getLogger(__name__)

DEFAULT_BEARER_TOKEN_HOURS = 5


class BaseAccount(object):
    def _build_canonical_str(self, url_components, req):
        # Build signing string
        lines = [req.method, ]
        headers_to_sign = dict()

        canonical_resource = url_components.path
        params = dict()
        if url_components.query:
            params_list = sorted(parse_qsl(url_components.query, True),
                                 key=lambda it: it[0])
            assert len(params_list) == len(set(it[0] for it in params_list))
            params = dict(params_list)
            convert = lambda kv: kv if kv[1] != '' else (kv[0], )
            params_str = '&'.join(['='.join(convert(kv)) for kv in params_list])

            canonical_resource = '%s?%s' % (canonical_resource, params_str)

        headers = req.headers
        logger.debug('headers before signing: %s', headers)
        for k, v in six.iteritems(headers):
            k = k.lower()
            if k in ('content-type', 'content-md5') or k.startswith('x-odps'):
                headers_to_sign[k] = v
        for k in ('content-type', 'content-md5'):
            if k not in headers_to_sign:
                headers_to_sign[k] = ''
        date_str = headers.get('Date')
        if not date_str:
            req_date = utils.formatdate(usegmt=True)
            headers['Date'] = req_date
            date_str = req_date
        headers_to_sign['date'] = date_str
        for param_key, param_value in six.iteritems(params):
            if param_key.startswith('x-odps-'):
                headers_to_sign[param_key] = param_value

        headers_to_sign = OrderedDict([(k, headers_to_sign[k])
                                       for k in sorted(headers_to_sign)])
        logger.debug('headers to sign: %s', headers_to_sign)
        for k, v in six.iteritems(headers_to_sign):
            if k.startswith('x-odps-'):
                lines.append('%s:%s' % (k, v))
            else:
                lines.append(v)

        lines.append(canonical_resource)
        return '\n'.join(lines)

    def sign_request(self, req, endpoint, region_name=None):
        raise NotImplementedError


class AliyunAccount(BaseAccount):
    """
    Account of aliyun.com
    """
    def __init__(self, access_id, secret_access_key):
        self.access_id = access_id
        self.secret_access_key = secret_access_key
        self._last_signature_date = None
        self._last_signature_key = None

    def _get_v4_signature_key(self, date_str, region_name):
        if date_str == self._last_signature_date:
            return self._last_signature_key

        k_secret = utils.to_binary("aliyun_v4" + self.secret_access_key)
        k_date = hmac.new(k_secret, utils.to_binary(date_str), hashlib.sha256).digest()
        k_region = hmac.new(k_date, utils.to_binary(region_name), hashlib.sha256).digest()
        k_service = hmac.new(k_region, b"odps", hashlib.sha256).digest()

        self._last_signature_date = date_str
        self._last_signature_key = hmac.new(k_service, b"aliyun_v4_request", hashlib.sha256).digest()
        return self._last_signature_key

    def calc_auth_str(self, canonical_str, region_name=None):
        if region_name is None:
            # use legacy v2 sign
            signature = base64.b64encode(hmac.new(
                utils.to_binary(self.secret_access_key), utils.to_binary(canonical_str),
                hashlib.sha1).digest())
            return 'ODPS %s:%s' % (self.access_id, utils.to_str(signature))
        else:
            # use v4 sign
            date_str = datetime.strftime(datetime.utcnow(), "%Y%m%d")
            credential = "/".join([self.access_id, date_str, region_name, "odps/aliyun_v4_request"])
            sign_key = self._get_v4_signature_key(date_str, region_name)
            signature = base64.b64encode(hmac.new(sign_key, utils.to_binary(canonical_str), hashlib.sha1).digest())
            return 'ODPS %s:%s' % (credential, utils.to_str(signature))

    def sign_request(self, req, endpoint, region_name=None):
        url = req.url[len(endpoint):]
        url_components = urlparse(unquote(url), allow_fragments=False)

        canonical_str = self._build_canonical_str(url_components, req)
        logger.debug('canonical string: %s', canonical_str)

        req.headers['Authorization'] = self.calc_auth_str(canonical_str, region_name)
        logger.debug('headers after signing: %r', req.headers)


class StsAccount(AliyunAccount):
    """
    Account of sts
    """
    def __init__(self, access_id, secret_access_key, sts_token):
        super(StsAccount, self).__init__(access_id, secret_access_key)
        self.sts_token = sts_token

    def sign_request(self, req, endpoint, region_name=None):
        super(StsAccount, self).sign_request(req, endpoint, region_name=region_name)
        if self.sts_token:
            req.headers['authorization-sts-token'] = self.sts_token


class AppAccount(BaseAccount):
    """
    Account for applications.
    """
    def __init__(self, access_id, secret_access_key):
        self.access_id = access_id
        self.secret_access_key = secret_access_key

    def sign_request(self, req, endpoint, region_name=None):
        auth_str = req.headers['Authorization']
        signature = base64.b64encode(hmac.new(
            utils.to_binary(self.secret_access_key), utils.to_binary(auth_str),
            hashlib.sha1).digest())
        app_auth_str = "account_provider:%s,signature_method:%s,access_id:%s,signature:%s" % (
            'aliyun', 'hmac-sha1', self.access_id, utils.to_str(signature))
        req.headers['application-authentication'] = app_auth_str
        logger.debug('headers after app signing: %r', req.headers)


class SignServer(object):
    class SignServerHandler(six.moves.BaseHTTPServer.BaseHTTPRequestHandler):
        def do_GET(self):
            self.send_response(200)
            self.send_header("Content-type", "text/plain")
            self.end_headers()
            self.wfile.write(b"PyODPS Account Server")

        def do_POST(self):
            try:
                self._do_POST()
            except:
                logger.exception("Failed to sign request on SignServer.")
                self.send_response(500)
                self.end_headers()

        def _do_POST(self):
            ctype, pdict = cgi.parse_header(self.headers.get('content-type'))
            if ctype == 'multipart/form-data':
                postvars = cgi.parse_multipart(self.rfile, pdict)
            elif ctype == 'application/x-www-form-urlencoded':
                length = int(self.headers.get('content-length'))
                postvars = six.moves.urllib.parse.parse_qs(self.rfile.read(length), keep_blank_values=1)
            else:
                self.send_response(400)
                self.end_headers()
                return

            self._sign(postvars)

        def _sign(self, postvars):
            if self.server._token is not None:
                auth = self.headers.get('Authorization')
                if not auth:
                    self.send_response(401)
                    self.end_headers()
                    return
                method, content = auth.split(' ', 1)
                method = method.lower()
                if method == 'token':
                    if content != self.server._token:
                        self.send_response(401)
                        self.end_headers()
                        return
                else:
                    self.send_response(401)
                    self.end_headers()
                    return

            assert len(postvars[b'access_id']) == 1 and len(postvars[b'canonical']) == 1
            access_id = utils.to_str(postvars[b'access_id'][0])
            canonical = utils.to_str(postvars[b'canonical'][0])
            if b'region_name' not in postvars:
                region_name = None
            else:
                region_name = utils.to_str(postvars[b'region_name'][0])
            secret_access_key = self.server._accounts[access_id]

            account = AliyunAccount(access_id, secret_access_key)
            auth_str = account.calc_auth_str(canonical, region_name)

            self.send_response(200)
            self.send_header("Content-Type", "text/json")
            self.end_headers()
            self.wfile.write(utils.to_binary(auth_str))

        def log_message(self, *args):
            return

    class SignServerCore(six.moves.socketserver.ThreadingMixIn, six.moves.BaseHTTPServer.HTTPServer):
        def __init__(self, *args, **kwargs):
            self._accounts = kwargs.pop('accounts', {})
            self._token = kwargs.pop('token', None)
            self._ready = False
            six.moves.BaseHTTPServer.HTTPServer.__init__(self, *args, **kwargs)
            self._ready = True

        def stop(self):
            self.shutdown()
            self.server_close()

    def __init__(self, token=None):
        self._server = None
        self._accounts = dict()
        self._token = token

    @property
    def server(self):
        return self._server

    @property
    def accounts(self):
        return self._accounts

    @property
    def token(self):
        return self._token

    def start(self, endpoint):
        def starter():
            self._server = self.SignServerCore(endpoint, self.SignServerHandler,
                                               accounts=self.accounts, token=self.token)
            self._server.serve_forever()

        thread = threading.Thread(target=starter)
        thread.daemon = True
        thread.start()
        while self._server is None or not self._server._ready:
            time.sleep(0.05)

    def stop(self):
        self._server.stop()


class SignServerError(Exception):
    def __init__(self, msg, code, content):
        super(SignServerError, self).__init__(msg)
        self.code = code
        self.content = content


class SignServerAccount(BaseAccount):
    _session_local = threading.local()

    def __init__(self, access_id, sign_endpoint=None, server=None, port=None, token=None):
        self.access_id = access_id
        self.sign_endpoint = sign_endpoint or (server, port)
        self.token = token

    @property
    def session(self):
        if not hasattr(type(self)._session_local, '_session'):
            adapter_options = dict(
                pool_connections=options.pool_connections,
                pool_maxsize=options.pool_maxsize,
                max_retries=options.retry_times,
            )
            session = requests.Session()
            # mount adapters with retry times
            session.mount(
                'http://', requests.adapters.HTTPAdapter(**adapter_options))
            session.mount(
                'https://', requests.adapters.HTTPAdapter(**adapter_options))

            self._session_local._session = session
        return self._session_local._session

    def sign_request(self, req, endpoint, region_name=None):
        url = req.url[len(endpoint):]
        url_components = urlparse(unquote(url), allow_fragments=False)

        canonical_str = self._build_canonical_str(url_components, req)
        logger.debug('canonical string: %s', canonical_str)

        headers = dict()
        if self.token:
            headers['Authorization'] = 'token ' + self.token

        sign_content = dict(access_id=self.access_id, canonical=canonical_str)
        if region_name is not None:
            sign_content["region_name"] = region_name
        resp = self.session.request(
            'post', 'http://%s:%s' % self.sign_endpoint, headers=headers, data=sign_content
        )
        if resp.status_code < 400:
            req.headers['Authorization'] = resp.text
            logger.debug('headers after signing: %r', req.headers)
        else:
            try:
                err_msg = resp_err = resp.text
            except:
                resp_err = resp.content
                err_msg = repr(resp_err)

            raise SignServerError(
                'Sign server returned error code: %d\n%s' % (resp.status_code, err_msg),
                resp.status_code,
                resp_err,
            )


class BearerTokenAccount(BaseAccount):
    def __init__(
        self, token=None, expired_hours=DEFAULT_BEARER_TOKEN_HOURS, get_bearer_token_fun=None
    ):
        self._get_bearer_token = get_bearer_token_fun or self.get_default_bearer_token
        self._token = token or self._get_bearer_token()
        self._reload_bearer_token_time()

        self._expired_time = timedelta(hours=expired_hours)

    @classmethod
    def from_environments(cls):
        expired_hours = int(os.getenv('ODPS_BEARER_TOKEN_HOURS', str(DEFAULT_BEARER_TOKEN_HOURS)))
        kwargs = {"expired_hours": expired_hours}
        if 'ODPS_BEARER_TOKEN' in os.environ:
            return cls(os.environ['ODPS_BEARER_TOKEN'], **kwargs)
        elif 'ODPS_BEARER_TOKEN_FILE' in os.environ:
            return cls(**kwargs)
        return None

    @staticmethod
    def get_default_bearer_token():
        token_file_name = os.getenv("ODPS_BEARER_TOKEN_FILE")
        if token_file_name and os.path.exists(token_file_name):
            with open(token_file_name, "r") as token_file:
                return token_file.read().strip()

        from cupid.runtime import context, RuntimeContext

        if not RuntimeContext.is_context_ready():
            return
        cupid_context = context()
        return cupid_context.get_bearer_token()

    def get_bearer_token_and_timestamp(self):
        self._check_bearer_token()
        return self._token, self._last_modified_time.timestamp()

    def _reload_bearer_token_time(self):
        if "ODPS_BEARER_TOKEN_TIMESTAMP_FILE" in os.environ:
            with open(os.getenv("ODPS_BEARER_TOKEN_TIMESTAMP_FILE"), "r") as ts_file:
                self._last_modified_time = datetime.fromtimestamp(float(ts_file.read()))
        else:
            self._last_modified_time = datetime.now()

    def _check_bearer_token(self):
        t = datetime.now()
        if self._last_modified_time is None:
            token = self._get_bearer_token()
            if token is None:
                return
            if token != self._token:
                self._token = token
                self._reload_bearer_token_time()
        elif (t - self._last_modified_time) > self._expired_time:
            token = self._get_bearer_token()
            if token is None:
                return
            self._token = token
            self._reload_bearer_token_time()

    @property
    def token(self):
        return self._token

    def sign_request(self, req, endpoint, region_name=None):
        self._check_bearer_token()
        url = req.url[len(endpoint):]
        url_components = urlparse(unquote(url), allow_fragments=False)
        self._build_canonical_str(url_components, req)
        req.headers['x-odps-bearer-token'] = self._token
        logger.debug('headers after signing: %r', req.headers)


class CredentialProviderAccount(StsAccount):
    def __init__(self, credential_provider):
        self.provider = credential_provider
        super(CredentialProviderAccount, self).__init__(None, None, None)

    def sign_request(self, req, endpoint, region_name=None):
        credential = self.provider.get_credentials()

        self.access_id = credential.get_access_key_id()
        self.secret_access_key = credential.get_access_key_secret()
        self.sts_token = credential.get_security_token()
        return super(CredentialProviderAccount, self).sign_request(
            req, endpoint, region_name=region_name
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import email.header
import inspect
import json
import re
from collections import OrderedDict
from xml.dom import minidom

import requests

from . import utils
from .compat import BytesIO, ElementTree, six


def _route_xml_path(root, *keys, **kw):
    create_if_not_exists = kw.get('create_if_not_exists', False)

    if isinstance(root, six.string_types):
        root = ElementTree.fromstring(root)

    for key in keys:
        if key == '.':
            return root
        prev = root
        root = root.find(key)
        if root is None:
            if not create_if_not_exists:
                return
            root = ElementTree.Element(key)
            prev.append(root)

    return root


def _extract_encoded_json(content):
    if not content.startswith("=?"):
        return json.loads(content)
    content, encoding = email.header.decode_header(content)[0]
    return json.loads(content.decode(encoding))


def _route_json_path(root, *keys, **kw):
    create_if_not_exists = kw.get('create_if_not_exists', False)

    if isinstance(root, six.string_types):
        root = _extract_encoded_json(root)

    for key in keys:
        prev = root
        root = root.get(key)
        if root is None:
            if not create_if_not_exists:
                return
            root = prev[key] = OrderedDict()

    return root


def parse_ndarray(array):
    try:
        import numpy as np
        return np.asarray(array)
    except ImportError:
        return array


def serialize_ndarray(array):
    try:
        return array.tolist()
    except AttributeError:
        return array


def _wrap_with_none(func):
    def new_func(x):
        if x is None:
            return x
        return func(x)

    return new_func


_serialize_types = dict()
_serialize_types['bool'] = (_wrap_with_none(utils.str_to_bool), _wrap_with_none(utils.bool_to_str))
_serialize_types['json'] = (_wrap_with_none(json.loads), _wrap_with_none(json.dumps))
_serialize_types['rfc822'] = (_wrap_with_none(utils.parse_rfc822), _wrap_with_none(utils.gen_rfc822))
_serialize_types['rfc822l'] = (
    _wrap_with_none(utils.parse_rfc822), _wrap_with_none(lambda s: utils.gen_rfc822(s, localtime=True))
)
_serialize_types['ndarray'] = (_wrap_with_none(parse_ndarray), _wrap_with_none(serialize_ndarray))


class SerializeField(object):
    def __init__(self, *keys, **kwargs):
        self._path_keys = keys

        self._required = kwargs.get('required', False)  # used when serialized
        self._blank_if_null = kwargs.get('blank_if_null',
                                         True if self._required else False)
        self._default = kwargs.get('default')
        if 'type' in kwargs:
            self._parse_callback, self._serialize_callback = _serialize_types[kwargs.pop('type')]
        else:
            self._parse_callback = kwargs.get('parse_callback')
            self._serialize_callback = kwargs.get('serialize_callback')

        self.set_to_parent = kwargs.get('set_to_parent', False)

    def _to_str(self, val):
        if isinstance(val, six.string_types):
            return utils.to_str(val)
        return val

    def _set_default_keys(self, *keys):
        if not self._path_keys:
            self._path_keys = keys

    def parse(self, root, **kwargs):
        raise NotImplementedError

    def serialize(self, root, value):
        raise NotImplementedError


class HasSubModelField(SerializeField):
    def __init__(self, model, *args, **kwargs):
        if isinstance(model, six.string_types):
            self._model_cls = None
            self._model_str = model
        else:
            self._model_cls = model
            self._model_str = None

        super(HasSubModelField, self).__init__(*args, **kwargs)

    @property
    def _model(self):
        if self._model_cls is not None:
            return self._model_cls

        models = self._model_str.split('.')
        model_name = models[0]

        module = None
        for stack in inspect.stack():
            globs = stack[0].f_globals
            if model_name in globs:
                possible_module = globs[model_name]
                if inspect.isclass(possible_module) and \
                        issubclass(possible_module, SerializableModel):
                    module = possible_module
                    break

        if module is None:
            raise ValueError('Unknown model name: %s' % self._model_str)

        res = None
        for model in models[1:]:
            if res is None:
                res = getattr(module, model)
            else:
                res = getattr(res, model)

        self._model_cls = res
        return res


_default_name_maker = dict(capitalized=utils.underline_to_capitalized, raw=lambda v: v, camel=utils.underline_to_camel)


class SerializableModelMetaClass(type):
    def __new__(mcs, name, bases, kv):
        slots = []
        fields = dict()
        for base in bases:
            base_slots = list(getattr(base, '__slots__', []))
            if '__weakref__' in base_slots:
                base_slots.remove('__weakref__')
            slots.extend(base_slots)
            fields.update(getattr(base, '__fields', dict()))
        slots.extend(kv.get('__slots__', []))
        fields.update(kv.get('__fields', dict()))

        attrs = []
        parent_attrs = []
        def_name = kv.pop('_' + name + '__default_name', 'capitalized')
        for attr, field in (pair for pair in six.iteritems(kv) if not pair[0].startswith('__')):
            if inspect.isclass(field) and issubclass(field, SerializeField):
                field = field()
            if isinstance(field, SerializeField):
                field._set_default_keys(_default_name_maker[def_name](attr))
                if not field.set_to_parent:
                    slots.append(attr)
                    attrs.append(attr)
                if field.set_to_parent:
                    parent_attrs.append(attr)
                fields[attr] = field
        kv['_parent_attrs'] = set(parent_attrs)

        slots = tuple(OrderedDict.fromkeys(slots))

        slots_pos = dict([(v, k) for k, v in enumerate(slots)])
        fields = OrderedDict(
            sorted(six.iteritems(fields), key=lambda s: slots_pos.get(s[0], float('inf'))))

        for attr in attrs:
            if attr in kv:
                del kv[attr]

        slots = tuple(slot for slot in slots if slot not in kv)
        if len(slots) > 0:
            kv['__slots__'] = slots
        if len(fields) > 0:
            kv['__fields'] = fields

        return type.__new__(mcs, name, bases, kv)


class SerializableModel(six.with_metaclass(SerializableModelMetaClass)):
    __slots__ = '_parent', '__weakref__'

    def __init__(self, **kwargs):
        slots = getattr(self, '__slots__', [])

        for k, v in six.iteritems(kwargs):
            if k in slots:
                setattr(self, k, v)
        for attr in slots:
            try:
                super(SerializableModel, self).__getattribute__(attr)
            except AttributeError:
                setattr(self, attr, None)

    @property
    def parent(self):
        return self._parent

    @classmethod
    def _is_null(cls, v):
        if v is None:
            return True
        if isinstance(v, (list, dict)) and len(v) == 0:
            return True
        return False

    @classmethod
    def _setattr(cls, obj, k, v, skip_null=True):
        if cls._is_null(v) and object.__getattribute__(obj, k) is not None:
            if not skip_null:
                setattr(obj, k, v)
            return

        fields = getattr(type(obj), '__fields')
        if not isinstance(fields[k], HasSubModelField):
            setattr(obj, k, v)
        elif isinstance(v, list):
            setattr(obj, k, v)
        else:
            sub_obj = object.__getattribute__(obj, k)
            new_obj = v
            if sub_obj is None:
                setattr(obj, k, v)
                return
            sub_fields = getattr(new_obj, '__fields', {})
            for k in six.iterkeys(sub_fields):
                if sub_fields[k].set_to_parent is True:
                    continue
                cls._setattr(sub_obj, k, object.__getattribute__(new_obj, k),
                             skip_null=skip_null)

    @classmethod
    def _init_obj(cls, content, obj=None, **kw):
        fields = dict(getattr(cls, '__fields'))

        _type = getattr(cls, '_type_indicator', None)
        _name = 'name' if 'name' in fields else None
        if obj is None and (_type is not None or 'name' in fields):
            kwargs = dict(kw)

            for field in (_name, _type):
                if field is None:
                    continue
                typo = fields[field].parse(content, **kw)
                kwargs[field] = typo

            obj = cls(**kwargs)

        return obj or cls(**kw)

    @classmethod
    def deserial(cls, content, obj=None, **kw):
        obj = cls._init_obj(content, obj=obj, **kw)
        obj_type = type(obj)

        fields = dict(getattr(obj_type, '__fields'))

        if isinstance(content, six.string_types):
            if issubclass(obj_type, XMLSerializableModel):
                content = ElementTree.fromstring(content)
            else:
                content = _extract_encoded_json(content)

        parent_kw = dict()
        self_kw = dict()

        for attr, prop in six.iteritems(fields):
            if isinstance(prop, SerializeField):
                kwargs = dict(kw)
                if isinstance(prop, HasSubModelField):
                    kwargs['_parent'] = obj
                if not prop.set_to_parent:
                    self_kw[attr] = prop.parse(content, **kwargs)
                else:
                    parent_kw[attr] = prop.parse(content, **kwargs)

        for k, v in six.iteritems(self_kw):
            obj_type._setattr(obj, k, v, skip_null=getattr(obj_type, 'skip_null', True))

        if obj.parent is not None:
            for k, v in six.iteritems(parent_kw):
                # remember that do not use `hasattr` here
                try:
                    old_v = object.__getattribute__(obj.parent, k)
                except AttributeError:
                    continue
                if v is not None and old_v != v:
                    setattr(obj.parent, k, v)

        return obj

    def serial(self):
        if isinstance(self, XMLSerializableModel):
            assert self._root is not None
            root = ElementTree.Element(self._root)
        else:
            root = OrderedDict()

        for attr, prop in six.iteritems(getattr(self, '__fields')):
            if isinstance(prop, SerializeField):
                try:
                    prop.serialize(root, object.__getattribute__(self, attr))
                except NotImplementedError:
                    continue

        return root

    def extract(self, **base_kw):
        kwargs = base_kw.copy()
        for attr in self.__slots__:
            try:
                kwargs[attr] = object.__getattribute__(self, attr)
            except AttributeError:
                pass
        return kwargs


class XMLSerializableModel(SerializableModel):
    __slots__ = '_root',

    @classmethod
    def parse(cls, response, obj=None, **kw):
        if 'parent' in kw:
            kw['_parent'] = kw.pop('parent')
        if isinstance(response, requests.Response):
            # PY2 prefer bytes, while PY3 prefer str
            response = response.content.decode() if six.PY3 else response.content
        return cls.deserial(response, obj=obj, **kw)

    def serialize(self):
        root = self.serial()

        sio = BytesIO()
        ElementTree.ElementTree(root).write(sio, encoding="utf-8", xml_declaration=True)
        xml_content = sio.getvalue()

        prettified_xml = minidom.parseString(xml_content).toprettyxml(indent=' '*2, encoding='utf-8')
        prettified_xml = utils.to_text(prettified_xml, encoding='utf-8')

        cdata_re = re.compile(r'&lt;!\[CDATA\[.*\]\]&gt;', (re.M | re.S))
        for src_cdata in cdata_re.finditer(prettified_xml):
            src_cdata = src_cdata.group(0)
            dest_cdata = src_cdata.replace('&amp;', '&').replace('&lt;', '<'). \
                replace('&quot;', '"').replace('&gt;', '>')
            prettified_xml = prettified_xml.replace(src_cdata, dest_cdata)

        return prettified_xml.replace('&quot;', '"')


class JSONSerializableModel(SerializableModel):
    @classmethod
    def parse(cls, response, obj=None, **kw):
        if 'parent' in kw:
            kw['_parent'] = kw.pop('parent')
        if isinstance(response, requests.Response):
            # PY2 prefer bytes, while PY3 prefer str
            response = response.content.decode() if six.PY3 else response.content
        return cls.deserial(response, obj=obj, **kw)

    def serialize(self, **kwargs):
        root = self.serial()
        return json.dumps(root, **kwargs)


class XMLTagField(SerializeField):
    def parse(self, root, **kwargs):
        node = _route_xml_path(root, *self._path_keys)

        val = self._default
        if node is not None:
            val = node.tag

        if val is None:
            return

        if self._parse_callback:
            return self._parse_callback(val)
        return val

    def _set_default_keys(self, *keys):
        super(XMLTagField, self)._set_default_keys('.')


class XMLNodeField(SerializeField):
    def parse(self, root, **kwargs):
        node = _route_xml_path(root, *self._path_keys)

        val = self._default
        if node is not None:
            val = node.text or self._default

        if val is None:
            return

        if self._parse_callback:
            return self._parse_callback(val)
        return val

    def serialize(self, root, value):
        value = value if value is not None else self._default
        if value is None and self._blank_if_null:
            value = ''

        if not self._required and value is None:
            return

        node = _route_xml_path(root, create_if_not_exists=True, *self._path_keys)
        if self._serialize_callback:
            node.text = utils.to_text(self._serialize_callback(value))
        else:
            node.text = utils.to_text(value)


class XMLNodeAttributeField(SerializeField):
    def __init__(self, *keys, **kwargs):
        self._attr = kwargs.pop('attr', None)

        super(XMLNodeAttributeField, self).__init__(*keys, **kwargs)

    def parse(self, root, **kwargs):
        assert self._attr is not None

        node = _route_xml_path(root, *self._path_keys)

        val = self._default
        if node is not None:
            val = node.get(self._attr)

        if val is None:
            return

        if self._parse_callback:
            return self._parse_callback(val)
        return node.get(self._attr)

    def serialize(self, root, value):
        assert self._attr is not None

        value = value if value is not None else self._default
        if value is None:
            if self._default is not None:
                value = self._default
            elif self._blank_if_null:
                value = ''

        if not self._required and value is None:
            return

        node = _route_xml_path(root, create_if_not_exists=True, *self._path_keys)
        if self._serialize_callback:
            node.set(self._attr, self._serialize_callback(value))
        else:
            node.set(self._attr, value)

    def _set_default_keys(self, *keys):
        if self._attr is None:
            self._attr = keys[0]


class XMLNodesField(SerializeField):
    def parse(self, root, **kwargs):
        prev_path_keys = self._path_keys[:-1]
        if prev_path_keys:
            root = _route_xml_path(root, *prev_path_keys)

        values = self._default
        if root is not None:
            values = [node.text for node in root.findall(self._path_keys[-1])]

        if values is None:
            return

        if self._parse_callback:
            return self._parse_callback(values)
        return values

    def serialize(self, root, value):
        value = value if value is not None else self._default
        if value is None and self._blank_if_null:
            value = []

        if not self._required and value is None:
            return

        if self._serialize_callback:
            value = self._serialize_callback(value)

        prev_path_keys = self._path_keys[:-1]
        if prev_path_keys:
            root = _route_xml_path(root, create_if_not_exists=True, *prev_path_keys)

        for val in value:
            element = ElementTree.Element(self._path_keys[-1])
            element.text = utils.to_text(val)
            root.append(element)


class XMLNodeReferenceField(HasSubModelField):
    def parse(self, root, **kwargs):
        node = _route_xml_path(root, *self._path_keys)

        instance = self._default
        if node is not None:
            if issubclass(self._model, JSONSerializableModel):
                node = node.text

            instance = self._model.deserial(node, **kwargs)
            if isinstance(instance, XMLSerializableModel) and \
                            instance._root is None:
                instance._root = node.tag

        if instance is None:
            return

        if self._parse_callback:
            return self._parse_callback(instance)
        return instance

    def serialize(self, root, value):
        value = value if value is not None else self._default

        if not self._required and value is None:
            return

        if self._serialize_callback:
            value = self._serialize_callback(value)

        prev_path_keys = self._path_keys[:-1]
        if prev_path_keys:
            root = _route_xml_path(root, create_if_not_exists=True, *prev_path_keys)

        if isinstance(value, XMLSerializableModel) and getattr(value, '_root') is None:
            setattr(value, '_root', self._path_keys[-1])

        val = value.serial()
        if isinstance(value, JSONSerializableModel):  # JSON mixed in XML
            element = ElementTree.Element(self._path_keys[-1])
            val = json.dumps(val)
            element.text = val
            root.append(element)
        else:
            root.append(val)


class XMLNodesReferencesField(HasSubModelField):
    def parse(self, root, **kwargs):
        prev_path_keys = self._path_keys[:-1]
        if prev_path_keys:
            root = _route_xml_path(root, *prev_path_keys)

        instances = self._default

        if root is not None:
            instances = []

            tag = self._path_keys[-1]
            if tag == '*':
                nodes = list(root)
            else:
                nodes = root.findall(self._path_keys[-1])

            for node in nodes:
                instance = self._model.deserial(node, **kwargs)
                if isinstance(instance, XMLSerializableModel) and instance._root is None:
                    instance._root = node.tag
                instances.append(instance)

        if instances is None:
            return

        if self._parse_callback:
            return self._parse_callback(instances)
        return instances

    def serialize(self, root, value):
        value = value if value is not None else self._default
        if value is None and self._blank_if_null:
            value = []

        if not self._required and value is None:
            return

        if self._serialize_callback:
            value = self._serialize_callback(value)

        prev_path_keys = self._path_keys[:-1]
        if prev_path_keys:
            root = _route_xml_path(root, create_if_not_exists=True, *prev_path_keys)

        for it in value:
            if isinstance(it, XMLSerializableModel) and \
                            getattr(it, '_root') is None:
                setattr(it, '_root', self._path_keys[-1])

            val = it.serial()
            if isinstance(it, JSONSerializableModel):
                element = ElementTree.Element(self._path_keys[-1])
                val = json.dumps(val)
                element.text = val
                root.append(element)
            else:
                root.append(val)


class XMLNodePropertiesField(SerializeField):
    def __init__(self, *keys, **kwargs):
        super(XMLNodePropertiesField, self).__init__(*keys, **kwargs)
        self._key_tag = kwargs['key_tag']
        self._value_tag = kwargs['value_tag']

    def parse(self, root, **kwargs):
        prev_path_keys = self._path_keys[:-1]
        if prev_path_keys:
            root = _route_xml_path(root, *prev_path_keys)

        results = self._default

        if root is not None:
            results = OrderedDict()

            for node in root.findall(self._path_keys[-1]):
                key_node = node.find(self._key_tag)
                value_node = node.find(self._value_tag)
                if key_node is not None and value_node is not None:
                    results[key_node.text] = value_node.text

        if results is None:
            return

        if self._parse_callback:
            return self._parse_callback(results)
        return results

    def serialize(self, root, value):
        value = value if value is not None else self._default
        if value is None and self._blank_if_null:
            value = OrderedDict()

        if not self._required and value is None:
            return

        if self._serialize_callback:
            value = self._serialize_callback(value)

        prev_path_keys = self._path_keys[:-1]
        if prev_path_keys:
            root = _route_xml_path(root, create_if_not_exists=True, *prev_path_keys)

        for k, v in six.iteritems(value):
            element = ElementTree.Element(self._path_keys[-1])
            key_node = ElementTree.Element(self._key_tag)
            key_node.text = utils.to_text(k)
            element.append(key_node)
            value_node = ElementTree.Element(self._value_tag)
            value_node.text = utils.to_text(v)
            element.append(value_node)

            root.append(element)


class JSONNodeField(SerializeField):
    def parse(self, root, **kwargs):
        val = self._default
        if root is not None:
            val = _route_json_path(root, *self._path_keys)

        if val is None:
            return

        val = self._to_str(val)
        if self._parse_callback:
            return self._parse_callback(val)
        return val

    def serialize(self, root, value):
        value = value if value is not None else self._default
        if value is None and self._blank_if_null:
            value = ''

        if not self._required and value is None:
            return

        prev_path_keys = self._path_keys[:-1]
        if prev_path_keys:
            node = _route_json_path(root, create_if_not_exists=True, *prev_path_keys)
        else:
            node = root

        if self._serialize_callback:
            node[self._path_keys[-1]] = self._serialize_callback(value)
        else:
            node[self._path_keys[-1]] = value


class JSONNodesField(SerializeField):
    def parse(self, root, **kwargs):
        prev_path_keys = self._path_keys[:-1]
        if prev_path_keys:
            root = _route_json_path(root, *prev_path_keys)

        values = self._default
        if root is not None:
            values = [self._to_str(node[self._path_keys[-1]])
                      for node in root]

        if values is None:
            return

        if self._parse_callback:
            return self._parse_callback(values)
        return values

    def serialize(self, root, value):
        value = value if value is not None else self._default
        if value is None and self._blank_if_null:
            value = []

        if not self._required and value is None:
            return

        if self._serialize_callback:
            value = self._serialize_callback(value)

        assert len(self._path_keys) >= 2

        prev_path_keys = self._path_keys[:-2]
        if prev_path_keys:
            root = _route_json_path(root, create_if_not_exists=True, *prev_path_keys)

        node = self._path_keys[-2]
        if node not in root:
            root[node] = []
        root = root[node]

        for val in value:
            root.append({self._path_keys[-1]: val})


class JSONNodeReferenceField(HasSubModelField):
    def __init__(self, model, *keys, **kwargs):
        super(JSONNodeReferenceField, self).__init__(model, *keys, **kwargs)

        self._check_before = kwargs.get('check_before')

    def parse(self, root, **kwargs):
        instance = self._default

        if root is not None:
            node = _route_json_path(root, *self._path_keys)

            if self._check_before is not None:
                _check = _route_json_path(root, *self._check_before)
                if not _check:
                    return

            instance = self._model.deserial(node, **kwargs)

        if instance is None:
            return

        if self._parse_callback:
            return self._parse_callback(instance)
        return instance

    def serialize(self, root, value):
        value = value if value is not None else self._default

        if not self._required and value is None:
            return

        if self._serialize_callback:
            value = self._serialize_callback(value)

        prev_path_keys = self._path_keys[:-1]
        if prev_path_keys:
            root = _route_json_path(root, create_if_not_exists=True, *prev_path_keys)

        root[self._path_keys[-1]] = value.serial()


class JSONNodesReferencesField(HasSubModelField):
    def parse(self, root, **kwargs):
        instances = self._default

        if isinstance(root, list):
            instances = [self._model.deserial(node, **kwargs)
                         for node in root]
        elif root is not None:
            prev_path_keys = self._path_keys[:-1]
            if prev_path_keys:
                root = _route_json_path(root, *prev_path_keys)

            if root is not None:
                root = root.get(self._path_keys[-1])
                if root is not None:
                    instances = [self._model.deserial(node, **kwargs)
                                 for node in root]

        if instances is None:
            return

        if self._parse_callback:
            return self._parse_callback(instances)
        return instances

    def serialize(self, root, value):
        value = value if value is not None else self._default
        if value is None and self._blank_if_null:
            value = []

        if not self._required and value is None:
            return

        if self._serialize_callback:
            value = self._serialize_callback(value)

        prev_path_keys = self._path_keys[:-1]
        if prev_path_keys:
            root = _route_json_path(root, create_if_not_exists=True, *prev_path_keys)

        key = self._path_keys[-1]
        if key not in root:
            root[key] = []
        [root[key].append(it.serial()) for it in value]


class JSONRawField(JSONNodeField):
    def _set_default_keys(self, *keys):
        return



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import json
import os
import socket
import subprocess
import sys
import time

import requests

from ....utils import to_str
from .client import CUPID_APP_NAME, NOTEBOOK_NAME


DEFAULT_NOTEBOOK_PORT = 50003
ACTOR_ADDRESS = "127.0.0.1:32123"
ACTOR_UID = "BearerTokenActor"


def start_notebook(port):
    proc = subprocess.Popen(
        [
            sys.executable,
            "-m",
            "jupyter",
            "notebook",
            "--ip",
            "0.0.0.0",
            "--no-browser",
            "--allow-root",
            "--port",
            str(port),
            "--NotebookApp.token=",
            "--NotebookApp.password=",
            "--NotebookApp.disable_check_xsrf=True",
            "--ServerApp.token=",
            "--ServerApp.password=",
            "--ServerApp.disable_check_xsrf=True",
        ]
    )
    while True:
        if proc.poll() is not None:
            raise SystemError("Notebook not started.")
        try:
            resp = requests.get("http://localhost:{}".format(port))
            if resp.status_code == 200:
                return
        except:
            pass
        time.sleep(1)


def wait_mars_ready(kvstore, name):
    # wait mars web ready
    while True:
        json_val = to_str(kvstore.get(name))
        if json_val:
            config = json.loads(to_str(json_val))
            mars_endpoint = to_str(config["endpoint"])
            return mars_endpoint


def dump_endpoint(endpoint):
    path = os.path.join(os.path.expanduser("~"), ".mars")
    with open(path, "w") as f:
        f.write(endpoint)


def refresh_bearer_token(cupid_context):
    path = os.path.join(os.path.expanduser("~"), ".bearertoken")
    token = cupid_context.get_bearer_token()

    refresh_time = 60
    while True:
        time.sleep(refresh_time)
        with open(path, "w") as f:
            f.write(token)


start_code = f"""
import asyncio
import os
from odps import options, ODPS
from odps.accounts import BearerTokenAccount

options.verbose = True

from mars.session import new_session
path = os.path.join(os.path.expanduser('~'), '.mars')
with open(path, 'r') as f:
    endpoint = f.read()
mars_session = new_session(endpoint).as_default()

def refresh_bearer_token():
    import mars.oscar as mo

    async def _refresher():
        ref = await mo.actor_ref(uid='{ACTOR_UID}', address='{ACTOR_ADDRESS}')
        return await ref.get_bearer_token()

    loop = mars_session._isolation.loop
    return asyncio.run_coroutine_threadsafe(_refresher(), loop).result()

bearer_token = refresh_bearer_token()
account = BearerTokenAccount(bearer_token, get_bearer_token_fun=refresh_bearer_token)
project = os.environ.get('ODPS_PROJECT_NAME')
endpoint = os.environ.get('ODPS_RUNTIME_ENDPOINT')
o = ODPS(None, None, account=account, project=project, endpoint=endpoint)
"""


def config_startup():
    startup_dir = os.path.join(
        os.path.expanduser("~"), ".ipython/profile_default/startup"
    )
    if not os.path.exists(startup_dir):
        os.makedirs(startup_dir)
    f_path = os.path.join(startup_dir, "set_default_session.py")
    with open(f_path, "w") as f:
        f.write(start_code)


async def create_bearer_token_actor():
    import mars.oscar as mo

    class BearerTokenActor(mo.Actor):
        def get_bearer_token(self):
            from cupid import context

            ctx = context()
            return ctx.get_bearer_token()

    pool = await mo.create_actor_pool(address=ACTOR_ADDRESS, n_process=1)
    await mo.create_actor(
        BearerTokenActor, uid=ACTOR_UID, address=pool.external_address
    )
    await pool.join()


def _main():
    from cupid import context
    from mars.utils import get_next_port

    cupid_context = context()
    mars_endpoint = wait_mars_ready(cupid_context.kv_store(), CUPID_APP_NAME)
    host_addr = socket.gethostbyname(socket.gethostname())

    os.environ.pop("KUBE_API_ADDRESS")

    if os.environ.get("VM_ENGINE_TYPE") == "hyper":
        notebook_port = DEFAULT_NOTEBOOK_PORT
    else:
        notebook_port = str(get_next_port())

    endpoint = "http://{0}:{1}".format(host_addr, notebook_port)

    # dump endpoint to ~/.mars
    dump_endpoint(mars_endpoint)

    # add startup script for notebook
    config_startup()

    # start notebook
    start_notebook(notebook_port)

    # modify in hyper mode
    if os.environ.get("VM_ENGINE_TYPE") == "hyper":
        endpoint = socket.gethostname() + "-{}".format(notebook_port)
    cupid_context.register_application(NOTEBOOK_NAME, endpoint)

    asyncio.run(create_bearer_token_actor())


if __name__ == "__main__":
    _main()



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class SubprocessStreamEOFError(IOError):
    pass


class CupidError(RuntimeError):
    pass


class InstanceRecycledError(CupidError):
    pass


class CupidUserError(CupidError):
    pass


class CupidMasterTimeoutError(CupidError):
    pass


class CupidCppError(CupidError):
    cpp_error_type = None

    def __new__(cls, err_type=None, err_msg=None):
        if cls is not CupidCppError:
            return CupidError.__new__(cls)
        if err_type in _cpp_errors:
            return CupidError.__new__(_cpp_errors[err_type])
        else:
            return CupidError.__new__(cls)

    def __init__(self, err_type=None, err_msg=None):
        self._err_type = err_type
        self._err_msg = err_msg

    def __str__(self):
        return '%s: %s' % (self._err_type, self._err_msg)


class CupidReplyTimeoutError(CupidCppError):
    cpp_error_type = 'ChannelTimeOutException'


class CupidChannelReplyError(CupidCppError):
    cpp_error_type = 'ChannelReplyException'


class CupidClientClosedError(CupidCppError):
    cpp_error_type = 'ChannelClientClosedException'


_cpp_errors = dict()
for _ex in globals().copy().values():
    if not isinstance(_ex, type) or not issubclass(_ex, CupidCppError):
        continue
    if getattr(_ex, 'cpp_error_type', None) is None:
        continue
    _cpp_errors[_ex.cpp_error_type] = _ex



from libc.stdint cimport *
from libcpp.vector cimport vector

from ..src.types_c cimport BaseRecord, SchemaSnapshot
from ..src.utils_c cimport CMillisecondsConverter


cdef class AbstractHasher:
    cdef int32_t c_hash_bigint(self, int64_t val) nogil
    cdef int32_t c_hash_float(self, float val) nogil
    cdef int32_t c_hash_double(self, double val) nogil
    cdef int32_t c_hash_bool(self, bint val) nogil
    cdef int32_t c_hash_string(self, char *ptr, size_t size) nogil


cdef class DefaultHasher(AbstractHasher):
    cdef int32_t c_hash_bigint(self, int64_t val) nogil
    cdef int32_t c_hash_float(self, float val) nogil
    cdef int32_t c_hash_double(self, double val) nogil
    cdef int32_t c_hash_bool(self, bint val) nogil
    cdef int32_t c_hash_string(self, char *ptr, size_t size) nogil


cdef class LegacyHasher(AbstractHasher):
    cdef int32_t c_hash_bigint(self, int64_t val) nogil
    cdef int32_t c_hash_float(self, float val) nogil
    cdef int32_t c_hash_double(self, double val) nogil
    cdef int32_t c_hash_bool(self, bint val) nogil
    cdef int32_t c_hash_string(self, char *ptr, size_t size) nogil


cpdef AbstractHasher get_hasher(hasher_type)


ctypedef int32_t (*_HASH_FUNCTION)(RecordHasher self, object value) except? -1


cdef class RecordHasher:
    cdef:
        AbstractHasher _hasher
        SchemaSnapshot _schema_snapshot
        vector[int32_t] _col_ids
        vector[_HASH_FUNCTION] _idx_to_hash_fun
        CMillisecondsConverter _mills_converter

    cdef int32_t _hash_bigint(self, object value) except? -1
    cdef int32_t _hash_float(self, object value) except? -1
    cdef int32_t _hash_double(self, object value) except? -1
    cdef int32_t _hash_bool(self, object value) except? -1
    cdef int32_t _hash_string(self, object value) except? -1
    cdef int32_t _hash_date(self, object value) except? -1
    cdef int32_t _hash_datetime(self, object value) except? -1
    cdef int32_t _hash_timestamp(self, object value) except? -1
    cdef int32_t _hash_timedelta(self, object value) except? -1
    cpdef int32_t hash(self, BaseRecord record)


cpdef int32_t hash_value(hasher_type, data_type, value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .core import RestModel
from .projects import Projects
from .project import Project
from .tables import Tables
from .table import Table, TableSchema
from .instances import Instances
from .instance import Instance
from .functions import Functions
from .function import Function
from .resources import Resources
from .resource import *
from .tenant import Tenant
from .volumes import *
from .volume_parted import PartedVolume, VolumePartition
from .volume_fs import FSVolume, FSVolumeDir, FSVolumeFile
from .volume_ext import ExternalVolume, ExternalVolumeDir, ExternalVolumeFile
from .xflows import XFlows
from .xflow import XFlow
from .tasks import *
from .record import Record
from .worker import Worker

import sys
import warnings

if sys.version_info[:2] < (3, 7):
    Schema = TableSchema  # Schema is to keep compatible
else:
    def __getattr__(name):
        if name != "Schema":
            raise AttributeError(name)

        from .. import utils

        warnings.warn(
            "Importing Schema from odps.models is deprecated, "
            "please use odps.models.TableSchema instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        utils.add_survey_call("odps.models.Schema")
        return TableSchema

Column = TableSchema.TableColumn
Partition = TableSchema.TablePartition

# keep renames compatible
VolumeFSDir = FSVolumeDir
VolumeFSFile = FSVolumeFile



# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

from ..expr.exporters import get_input_field_names
from ..expr.mixin import ml_collection_mixin
from ...compat import Enum


class TimeSeriesFieldRole(Enum):
    TS_GROUP = 'TS_GROUP'
    TS_SEQ = 'TS_SEQ'
    TS_VALUE = 'TS_VALUE'


@ml_collection_mixin
class TimeSeriesMLMixIn(object):
    __slots__ = ()

    field_role_enum = TimeSeriesFieldRole
    non_feature_roles = set((TimeSeriesFieldRole.TS_GROUP, TimeSeriesFieldRole.TS_SEQ))


"""
Common time series exporters
"""
get_ts_group_column = partial(get_input_field_names, field_role=TimeSeriesFieldRole.TS_GROUP)
get_ts_seq_column = partial(get_input_field_names, field_role=TimeSeriesFieldRole.TS_SEQ)
get_ts_value_column = partial(get_input_field_names, field_role=TimeSeriesFieldRole.TS_VALUE)



# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from ..df import RandomScalar, Delay
from ..compat import six
from ..utils import with_wait_argument
from .utils import get_function_args
from .metrics import SCORERS

try:
    import numpy as np
    _has_numpy = True
except ImportError:
    _has_numpy = False


def k_fold(n_folds=3):
    def _calc(df):
        seed = int(time.time())
        rand_col = '_rand_key_%s' % seed
        old_columns = df.schema.names
        rand_df = df[df, RandomScalar(seed).rename(rand_col)].cache()

        for idx in range(n_folds):
            left_df = rand_df[idx * 1.0 / n_folds < getattr(rand_df, rand_col) <= (idx + 1) * 1.0 / n_folds].__getitem__(old_columns)
            right_df = rand_df[(getattr(rand_df, rand_col) > (idx + 1) * 1.0 / n_folds) |
                               (getattr(rand_df, rand_col) <= idx * 1.0 / n_folds)].__getitem__(old_columns)
            yield (left_df, right_df)
    return _calc


def train_test(train_size=None, test_size=None):
    if train_size and test_size and train_size + test_size < 1.0e-10:
        raise ValueError('Sum of train size and test size must be 1.0.')
    if train_size and (train_size < 0 or train_size >= 1.0):
        raise ValueError('Illegal train size.')
    if test_size and (test_size < 0 or test_size >= 1.0):
        raise ValueError('Illegal train size.')
    if test_size:
        train_size = 1.0 - test_size
    if not (train_size or test_size):
        train_size = 0.75

    def _calc(df):
        return tuple(df.split(train_size))

    return _calc


@with_wait_argument
def cross_val_score(trainer, df, col_true=None, col_pred=None, col_score=None, scoring=None, cv=None,
                    fit_params=None, ui=None, async_=False, n_parallel=1, timeout=None,
                    close_and_notify=True):
    if cv is None:
        cv = k_fold()
    elif isinstance(cv, six.integer_types):
        cv = k_fold(cv)
    if scoring is None:
        scoring = SCORERS['accuracy']

    if fit_params:
        [setattr(trainer, param, val) for param, val in six.iteritems(fit_params)]

    kwargs = dict(col_true=col_true, col_pred=col_pred)
    if 'col_score' in get_function_args(scoring):
        kwargs['col_score'] = col_score

    delay = Delay()
    futures = []
    for train_df, test_df in cv(df):
        model = trainer.train(train_df)
        predicted = model.predict(test_df)
        metrics_expr = scoring(df=predicted, execute_now=False, **kwargs)
        futures.append(metrics_expr.execute(delay=delay))

    delay.execute(ui=ui, async_=async_, n_parallel=n_parallel, timeout=timeout, close_and_notify=close_and_notify)

    results = [f.result() for f in futures]
    if _has_numpy:
        return np.array(results)
    else:
        return results



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import json
import itertools

import numpy as np
from mars.tensor.datasource.core import TensorNoInput
from mars.tensor.utils import decide_chunk_sizes, normalize_shape
from mars.serialize import ValueType, ListField, StringField
from mars import opcodes as OperandDef

from ..io import glob
from ....compat import izip, BytesIO

logger = logging.getLogger(__name__)


class TensorTableCOO(TensorNoInput):
    _op_type_ = OperandDef.TABLE_COO

    _paths = ListField("paths", ValueType.string)
    _dim_cols = ListField("dim_cols", ValueType.string)
    _value_col = StringField("value_col")
    _storage_options = StringField("storage_options")

    def __init__(
        self,
        dtype=None,
        paths=None,
        dim_cols=None,
        value_col=None,
        storage_options=None,
        sparse=True,
        **kw
    ):
        super(TensorTableCOO, self).__init__(
            _paths=paths,
            _dim_cols=dim_cols,
            _value_col=value_col,
            _dtype=dtype,
            _storage_options=storage_options,
            _sparse=sparse,
            **kw
        )

    @property
    def paths(self):
        return self._paths

    @property
    def dim_cols(self):
        return self._dim_cols

    @property
    def value_col(self):
        return self._value_col

    @property
    def storage_options(self):
        return self._storage_options

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]

        storage_opts = json.loads(op.storage_options)

        logger.debug("Start scanning data files in %s", op.paths[0])
        chunk_files = dict()
        for key in glob(op.paths[0], **storage_opts):
            file_name, _ = key.rsplit(".", 1)
            _, fn_suffix = file_name.rsplit("/", 1)
            dim_suffix = fn_suffix.rsplit("@", 1)[-1]
            dim_indices = tuple(int(pt) for pt in dim_suffix.split(","))
            if dim_indices not in chunk_files:
                chunk_files[dim_indices] = []
            chunk_files[dim_indices].append(key)
        logger.debug("Finish scanning data files in %s", op.paths[0])

        try:
            target_chunk_size = tensor.params.raw_chunk_size
        except AttributeError:
            target_chunk_size = tensor.extra_params.raw_chunk_size
        chunk_size = decide_chunk_sizes(
            tensor.shape, target_chunk_size, tensor.dtype.itemsize
        )
        chunk_size_idxes = (range(len(size)) for size in chunk_size)

        out_chunks = []
        for chunk_shape, chunk_idx in izip(
            itertools.product(*chunk_size), itertools.product(*chunk_size_idxes)
        ):
            chunk_op = op.copy().reset_key()
            chunk_op._paths = chunk_files.get(chunk_idx, [])
            out_chunk = chunk_op.new_chunk(None, shape=chunk_shape, index=chunk_idx)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs, tensor.shape, nsplits=chunk_size, chunks=out_chunks
        )

    @classmethod
    def execute(cls, ctx, op):
        import pyarrow.parquet as pq
        import pandas as pd
        import scipy.sparse as sps
        from mars.lib.sparse import SparseNDArray
        from ..io import open as fs_open

        dfs = []
        storage_opts = json.loads(op.storage_options)
        for p in op.paths:
            with fs_open(p, "rb", **storage_opts) as inp_file:
                f = inp_file.read()
                dfs.append(pq.read_table(BytesIO(f)).to_pandas())

        chunk = op.outputs[0]
        if op.sparse and len(dfs) == 0:
            if len(chunk.shape) == 1:
                csr_array = sps.csr_matrix((chunk.shape[0], 1))
                ctx[chunk.key] = SparseNDArray(csr_array, shape=chunk.shape)
            else:
                csr_array = sps.csr_matrix(chunk.shape)
                ctx[chunk.key] = SparseNDArray(csr_array)
            return

        df_merged = pd.concat(dfs, ignore_index=True)
        dim_arrays = [df_merged[col] for col in op.dim_cols]
        value_array = df_merged[op.value_col].astype(chunk.dtype)
        del df_merged

        if op.sparse:
            if len(chunk.shape) == 1:
                dim_arrays.append(np.zeros((len(dim_arrays[0]))))
                csr_array = sps.csr_matrix(
                    (value_array, tuple(dim_arrays)), shape=(chunk.shape[0], 1)
                )
            else:
                csr_array = sps.csr_matrix(
                    (value_array, tuple(dim_arrays)), shape=chunk.shape
                )
            del dim_arrays, value_array
            ctx[chunk.key] = SparseNDArray(csr_array, shape=chunk.shape)
        else:
            arr = np.empty(chunk.shape, dtype=value_array.dtype)
            arr[tuple(dim_arrays)] = value_array
            ctx[chunk.key] = arr


def read_coo(
    path_pattern,
    dim_cols,
    value_col,
    dtype=float,
    shape=None,
    chunk_size=None,
    sparse=False,
    **storage_opts
):
    if sparse and len(dim_cols) > 2:
        raise ValueError("Can only support reading 1-d or 2-d data if sparse")

    dtype = np.dtype(dtype)
    op = TensorTableCOO(
        dtype=dtype,
        paths=[path_pattern],
        dim_cols=dim_cols,
        value_col=value_col,
        storage_options=json.dumps(storage_opts),
        sparse=sparse,
    )
    return op(normalize_shape(shape), chunk_size=chunk_size)



# encoding: utf-8
# copyright 1999-2022 alibaba group holding ltd.
#
# licensed under the apache license, version 2.0 (the "license");
# you may not use this file except in compliance with the license.
# you may obtain a copy of the license at
#
#      http://www.apache.org/licenses/license-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the license is distributed on an "as is" basis,
# without warranties or conditions of any kind, either express or implied.
# see the license for the specific language governing permissions and
# limitations under the license.

import sys
import types

from .classification import *
from .regression import *
from .clustering import *
from .scorer import SCORERS

from ..algolib.loader import load_defined_algorithms

# required by autodoc of sphinx
__all__ = [k for k, v in globals().items()
           if not k.startswith('_') and isinstance(v, (types.FunctionType, type))]

load_defined_algorithms(sys.modules[__name__], 'metrics')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .groupby import *
from .reduction import _stats_type
from . import utils, errors
from .. import types
from ...utils import camel_to_underline, object_getattr


class Window(SequenceExpr):
    _args = '_input', '_partition_by', '_order_by', '_preceding', '_following'

    def _init(self, *args, **kwargs):
        self._preceding = None
        self._following = None

        super(Window, self)._init(*args, **kwargs)

        self._preceding, self._following = \
            self._de_scalar(self._preceding), self._de_scalar(self._following)

        if isinstance(self._preceding, list):
            self._preceding = tuple(self._preceding)
        elif isinstance(self._following, list):
            self._following = tuple(self._following)

        if (isinstance(self._preceding, tuple) and self._following is not None) or \
            (isinstance(self._following, tuple) and self._preceding is not None):
            raise ValueError('Cannot specify window of both sides')

        if isinstance(self._preceding, tuple):
            start, end = self._preceding
            if start is None:
                assert end >= 0
            else:
                assert start > end
        elif isinstance(self._following, tuple):
            start, end = self._following
            if end is None:
                assert start >= 0
            else:
                assert start < end
        else:
            if self._preceding is not None and self._preceding < 0:
                raise ValueError('Window offset must be positive')

            if self._following is not None and self._following < 0:
                raise ValueError('Window offset must be positive')
        self._preceding, self._following = \
            self._scalar(self._preceding), self._scalar(self._following)

    def _scalar(self, val):
        if val is None:
            return
        if isinstance(val, Scalar):
            return val
        if isinstance(val, tuple):
            return tuple(self._scalar(it) for it in val)
        else:
            return Scalar(_value=val)

    def _de_scalar(self, val):
        if val is None:
            return
        if isinstance(val, tuple):
            return tuple(self._de_scalar(it) for it in val)
        elif isinstance(val, Scalar):
            return val._value
        else:
            return val

    def iter_args(self):
        for it in zip(['Column', 'PartitionBy', 'OrderBy',
                       'preceding', 'following'], self.args):
            yield it

    @property
    def name(self):
        if self._name:
            return self._name
        source_name = self.source_name
        if source_name:
            name = self.node_name
            if name.startswith('Cum'):
                name = name[3:]
            return '%s_%s' % (source_name, name.lower())

    @property
    def source_name(self):
        input_name = None
        if isinstance(self._input, SequenceExpr):
            input_name = self._input.name
        return self._source_name or input_name

    @property
    def node_name(self):
        return self.__class__.__name__

    @property
    def input(self):
        return self._input

    @property
    def preceding(self):
        return self._de_scalar(self._preceding)

    @property
    def following(self):
        return self._de_scalar(self._following)


class CumulativeOp(Window):
    _args = '_input', '_partition_by', '_distinct', '_order_by', \
            '_preceding', '_following'

    def _init(self, *args, **kwargs):
        self._distinct = False

        super(CumulativeOp, self)._init(*args, **kwargs)

        self._distinct = self._de_scalar(self._distinct)
        self._check_unique()
        self._distinct = self._scalar(self._distinct)

    def _check_unique(self):
        if self._distinct and len(getattr(self, '_order_by', None) or []) > 0:
            raise errors.ExpressionError('Unique and sort cannot exist both')

    def unique(self):
        if self._distinct.value:
            return self

        self._check_unique()

        attr_values = dict((attr, getattr(self, attr, None))
                           for attr in utils.get_attrs(self))
        attr_values['_distinct'] = True
        return type(self)(**attr_values)

    def iter_args(self):
        for it in zip(['Column', 'PartitionBy', 'distinct', 'OrderBy',
                       'preceding', 'following'], self.args):
            yield it

    @property
    def distinct(self):
        return self._distinct.value

    def accept(self, visitor):
        return visitor.visit_cum_window(self)


class CumSum(CumulativeOp):
    __slots__ = ()


class CumMax(CumulativeOp):
    __slots__ = ()


class CumMin(CumulativeOp):
    __slots__ = ()


class CumMean(CumulativeOp):
    __slots__ = ()


class CumMedian(CumulativeOp):
    __slots__ = ()


class CumCount(CumulativeOp):
    __slots__ = ()


class CumStd(CumulativeOp):
    __slots__ = ()


class NthValue(CumulativeOp):
    __slots__ = '_nth', '_skip_nulls'


class RankOp(Window):
    __slots__ = ()

    @property
    def name(self):
        if self._name:
            return self._name
        return camel_to_underline(self.__class__.__name__)

    def accept(self, visitor):
        return visitor.visit_rank_window(self)


class Rank(RankOp):
    __slots__ = ()


class DenseRank(RankOp):
    __slots__ = ()


class PercentRank(RankOp):
    __slots__ = ()


class RowNumber(RankOp):
    __slots__ = ()


class QCut(RankOp):
    __slots__ = '_bins',


class CumeDist(RankOp):
    __slots__ = ()


class ShiftOp(Window):
    _args = '_input', '_partition_by', '_order_by', '_offset', '_default'

    def _init(self, *args, **kwargs):
        self._init_attr('_offset', None)
        self._init_attr('_default', None)
        super(ShiftOp, self)._init(*args, **kwargs)

        if self._offset is not None and not isinstance(self._offset, Expr):
            self._offset = self._scalar(self._offset)
        if self._default is not None and not isinstance(self._default, Expr):
            self._default = self._scalar(self._default)

    def accept(self, visitor):
        return visitor.visit_shift_window(self)


class Lag(ShiftOp):
    __slots__ = ()


class Lead(ShiftOp):
    __slots__ = ()


def _cumulative_op(expr, op_cls, sort=None, ascending=True, unique=False,
                   preceding=None, following=None, data_type=None, **kwargs):
    if isinstance(expr, SequenceGroupBy):
        if sort is not None:
            groupby = expr._input.sort(sort, ascending=ascending)
            expr = groupby[expr._name]

        collection = expr._input._input
        column = collection[expr._name]

        data_type = data_type or expr._data_type

        return op_cls(_input=column, _partition_by=expr._input._by,
                      _order_by=object_getattr(expr._input, '_sorted_fields', None),
                      _preceding=preceding, _following=following,
                      _data_type=data_type, _distinct=unique, **kwargs)


def cumsum(expr, sort=None, ascending=True, unique=False,
           preceding=None, following=None):
    """
    Calculate cumulative summation of a sequence expression.

    :param expr: expression for calculation
    :param sort: name of the sort column
    :param ascending: whether to sort in ascending order
    :param unique: whether to eliminate duplicate entries
    :param preceding: the start point of a window
    :param following: the end point of a window
    :return: calculated column
    """
    if expr._data_type == types.boolean:
        output_type = types.int64
    else:
        output_type = expr._data_type
    return _cumulative_op(expr, CumSum, sort=sort, ascending=ascending,
                          unique=unique, preceding=preceding,
                          following=following, data_type=output_type)


def cummax(expr, sort=None, ascending=True, unique=False,
           preceding=None, following=None):
    """
    Calculate cumulative maximum of a sequence expression.

    :param expr: expression for calculation
    :param sort: name of the sort column
    :param ascending: whether to sort in ascending order
    :param unique: whether to eliminate duplicate entries
    :param preceding: the start point of a window
    :param following: the end point of a window
    :return: calculated column
    """
    return _cumulative_op(expr, CumMax, sort=sort, ascending=ascending,
                          unique=unique, preceding=preceding,
                          following=following)


def cummin(expr, sort=None, ascending=True, unique=False,
           preceding=None, following=None):
    """
    Calculate cumulative minimum of a sequence expression.

    :param expr: expression for calculation
    :param sort: name of the sort column
    :param ascending: whether to sort in ascending order
    :param unique: whether to eliminate duplicate entries
    :param preceding: the start point of a window
    :param following: the end point of a window
    :return: calculated column
    """
    return _cumulative_op(expr, CumMin, sort=sort, ascending=ascending,
                          unique=unique, preceding=preceding,
                          following=following)


def cummean(expr, sort=None, ascending=True, unique=False,
            preceding=None, following=None):
    """
    Calculate cumulative mean of a sequence expression.

    :param expr: expression for calculation
    :param sort: name of the sort column
    :param ascending: whether to sort in ascending order
    :param unique: whether to eliminate duplicate entries
    :param preceding: the start point of a window
    :param following: the end point of a window
    :return: calculated column
    """
    data_type = _stats_type(expr)
    return _cumulative_op(expr, CumMean, sort=sort, ascending=ascending,
                          unique=unique, preceding=preceding,
                          following=following, data_type=data_type)


def cummedian(expr, sort=None, ascending=True, unique=False,
              preceding=None, following=None):
    """
    Calculate cumulative median of a sequence expression.

    :param expr: expression for calculation
    :param sort: name of the sort column
    :param ascending: whether to sort in ascending order
    :param unique: whether to eliminate duplicate entries
    :param preceding: the start point of a window
    :param following: the end point of a window
    :return: calculated column
    """
    data_type = _stats_type(expr)
    return _cumulative_op(expr, CumMedian, sort=sort, ascending=ascending,
                          unique=unique, preceding=preceding,
                          following=following, data_type=data_type)


def cumcount(expr, sort=None, ascending=True, unique=False,
             preceding=None, following=None):
    """
    Calculate cumulative count of a sequence expression.

    :param expr: expression for calculation
    :param sort: name of the sort column
    :param ascending: whether to sort in ascending order
    :param unique: whether to eliminate duplicate entries
    :param preceding: the start point of a window
    :param following: the end point of a window
    :return: calculated column
    """
    data_type = types.int64
    return _cumulative_op(expr, CumCount, sort=sort, ascending=ascending,
                          unique=unique, preceding=preceding,
                          following=following, data_type=data_type)


def cumstd(expr, sort=None, ascending=True, unique=False,
           preceding=None, following=None):
    """
    Calculate cumulative standard deviation of a sequence expression.

    :param expr: expression for calculation
    :param sort: name of the sort column
    :param ascending: whether to sort in ascending order
    :param unique: whether to eliminate duplicate entries
    :param preceding: the start point of a window
    :param following: the end point of a window
    :return: calculated column
    """
    data_type = _stats_type(expr)
    return _cumulative_op(expr, CumStd, sort=sort, ascending=ascending,
                          unique=unique, preceding=preceding,
                          following=following, data_type=data_type)


def nth_value(expr, nth, skip_nulls=False, sort=None, ascending=True):
    """
    Get nth value of a grouped and sorted expression.

    :param expr: expression for calculation
    :param nth: integer position
    :param skip_nulls: whether to skip null values, False by default
    :param sort: name of the sort column
    :param ascending: whether to sort in ascending order
    :return: calculated column
    """
    return _cumulative_op(expr, NthValue, data_type=expr._data_type, sort=sort,
                          ascending=ascending, _nth=nth, _skip_nulls=skip_nulls)


def _rank_op(expr, op_cls, data_type, sort=None, ascending=True, **kwargs):
    if isinstance(expr, SequenceGroupBy):
        grouped = expr._input
        sort = sort or object_getattr(grouped, '_sorted_fields', None) or expr.name
    elif not isinstance(expr, BaseGroupBy):
        raise NotImplementedError
    else:
        grouped = expr

    if not isinstance(grouped, SortedGroupBy) and sort is not None:
        grouped = grouped.sort(sort, ascending=ascending)
        if isinstance(expr, SequenceGroupBy):
            expr = grouped[expr.name]
    if not isinstance(grouped, SortedGroupBy):
        raise ExpressionError('`sort` arg is required for the rank operation')

    return op_cls(_input=grouped._input, _partition_by=grouped._by,
                  _order_by=object_getattr(grouped, '_sorted_fields', None),
                  _data_type=data_type, **kwargs)


def rank(expr, sort=None, ascending=True):
    """
    Calculate rank of a sequence expression.

    :param expr: expression for calculation
    :param sort: name of the sort column
    :param ascending: whether to sort in ascending order
    :return: calculated column
    """
    return _rank_op(expr, Rank, types.int64, sort=sort, ascending=ascending)


def dense_rank(expr, sort=None, ascending=True):
    """
    Calculate dense rank of a sequence expression.

    :param expr: expression for calculation
    :param sort: name of the sort column
    :param ascending: whether to sort in ascending order
    :return: calculated column
    """
    return _rank_op(expr, DenseRank, types.int64, sort=sort, ascending=ascending)


def percent_rank(expr, sort=None, ascending=True):
    """
    Calculate percentage rank of a sequence expression.

    :param expr: expression for calculation
    :param sort: name of the sort column
    :param ascending: whether to sort in ascending order
    :return: calculated column
    """
    return _rank_op(expr, PercentRank, types.float64, sort=sort, ascending=ascending)


def row_number(expr, sort=None, ascending=True):
    """
    Calculate row number of a sequence expression.

    :param expr: expression for calculation
    :param sort: name of the sort column
    :param ascending: whether to sort in ascending order
    :return: calculated column
    """
    return _rank_op(expr, RowNumber, types.int64, sort=sort, ascending=ascending)


def qcut(expr, bins, labels=False, sort=None, ascending=True):
    """
    Get quantile-based bin indices of every element of a grouped and sorted expression.
    The indices of bins start from 0. If cuts are not of equal sizes, extra items will
    be appended into the first group.

    :param expr: expression for calculation
    :param bins: number of bins
    :param sort: name of the sort column
    :param ascending: whether to sort in ascending order
    :return: calculated column
    """
    if labels is None or labels:
        raise NotImplementedError('Showing bins or customizing labels not supported')
    return _rank_op(expr, QCut, types.int64, sort=sort, ascending=ascending,
                    _bins=bins)


def cume_dist(expr, sort=None, ascending=True):
    """
    Calculate cumulative ratio of a sequence expression.

    :param expr: expression for calculation
    :param sort: name of the sort column
    :param ascending: whether to sort in ascending order
    :return: calculated column
    """
    return _rank_op(expr, CumeDist, types.float64, sort=sort, ascending=ascending)


def _shift_op(expr, op_cls, offset, default=None, sort=None, ascending=True):
    if isinstance(expr, SequenceGroupBy):
        if object_getattr(expr._input, '_sorted_fields', None) is None:
            sort = sort or expr.name
            groupby = expr._input.sort(sort, ascending=ascending)
            expr = groupby[expr._name]

        collection = expr._input._input
        column = collection[expr._name]

        return op_cls(_input=column, _partition_by=expr._input._by,
                      _order_by=object_getattr(expr._input, '_sorted_fields', None),
                      _offset=offset, _default=default,
                      _name=expr._name, _data_type=expr._data_type)


def lag(expr, offset, default=None, sort=None, ascending=True):
    """
    Get value in the row ``offset`` rows prior to the current row.

    :param offset: the offset value
    :param default: default value for the function, when there are no rows satisfying the offset
    :param expr: expression for calculation
    :param sort: name of the sort column
    :param ascending: whether to sort in ascending order
    :return: calculated column
    """
    return _shift_op(expr, Lag, offset, default=default,
                     sort=sort, ascending=ascending)


def lead(expr, offset, default=None, sort=None, ascending=True):
    """
    Get value in the row ``offset`` rows after to the current row.

    :param offset: the offset value
    :param default: default value for the function, when there are no rows satisfying the offset
    :param expr: expression for calculation
    :param sort: name of the sort column
    :param ascending: whether to sort in ascending order
    :return: calculated column
    """
    return _shift_op(expr, Lead, offset, default=default,
                     sort=sort, ascending=ascending)


_number_window_methods = dict(
    cumsum=cumsum,
    cummean=cummean,
    cummedian=cummedian,
    cumstd=cumstd,
)

_window_methods = dict(
    cummax=cummax,
    cummin=cummin,
    cumcount=cumcount,
    lag=lag,
    lead=lead,
)

_groupby_methods = dict(
    rank=rank,
    min_rank=rank,
    dense_rank=dense_rank,
    percent_rank=percent_rank,
    row_number=row_number,
    qcut=qcut,
    nth_value=nth_value,
    cume_dist=cume_dist,
)

number_windows = [globals().get(repr(t).capitalize() + SequenceGroupBy.__name__)
                  for t in types.number_types()]

for number_window in number_windows:
    utils.add_method(number_window, _number_window_methods)

utils.add_method(SequenceGroupBy, _window_methods)

StringSequenceGroupBy.cumsum = cumsum  # FIXME: should we support string?
BooleanSequenceGroupBy.cumsum = cumsum

utils.add_method(BaseGroupBy, _groupby_methods)
utils.add_method(SequenceGroupBy, _groupby_methods)



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libc.stdint cimport *
from libc.string cimport *
from ...src.stringstream cimport stringstream


cdef class CEncoder:
    cdef stringstream *_buffer

    cdef size_t position(self) nogil
    cpdef bytes tostring(self)
    cdef int append_tag(self, int field_num, int wire_type) except + nogil
    cdef int append_sint32(self, int32_t value) except + nogil
    cdef int append_uint32(self, uint32_t value) except + nogil
    cdef int append_sint64(self, int64_t value) except + nogil
    cdef int append_uint64(self, uint64_t value) except + nogil
    cdef int append_bool(self, bint value) except + nogil
    cdef int append_double(self, double value) except + nogil
    cdef int append_float(self, float value) except + nogil
    cdef int append_string(self, const char *ptr, size_t value_len) except + nogil


cdef class Encoder:
    cdef CEncoder _encoder



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..io.stream import RequestsIO, get_compress_stream
from ... import options
from ...compat import six

try:
    from .pdwriter_c import BasePandasWriter
except ImportError:
    BasePandasWriter = None


if BasePandasWriter:
    class TunnelPandasWriter(BasePandasWriter):
        def __init__(self, schema, request_callback, compress_option=None):
            self._req_io = RequestsIO(request_callback, chunk_size=options.chunk_size)
            out = get_compress_stream(self._req_io, compress_option)
            super(TunnelPandasWriter, self).__init__(schema, out)
            self._req_io.start()

        def write(self, data, columns=None, limit=-1, dim_offsets=None):
            if self._req_io._async_err:
                ex_type, ex_value, tb = self._req_io._async_err
                six.reraise(ex_type, ex_value, tb)
            super(TunnelPandasWriter, self).write(data, columns=columns, limit=limit,
                                                  dim_offsets=dim_offsets)

        write.__doc__ = BasePandasWriter.write.__doc__

        def close(self):
            super(TunnelPandasWriter, self).close()
            self._req_io.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


try:
    from .core import (
        create_mars_cluster,
        to_mars_dataframe,
        persist_mars_dataframe,
        run_script_in_mars,
        run_mars_job,
        list_mars_instances,
        sql_to_mars_dataframe,
    )
except ImportError:
    create_mars_cluster = None
    to_mars_dataframe = None
    persist_mars_dataframe = None
    run_mars_script = None
    run_mars_job = None
    list_mars_instances = None
    sql_to_mars_dataframe = None

try:
    from . import dataframe
except ImportError:
    dataframe = None

try:
    from . import tensor
except ImportError:
    tensor = None

try:
    from mars.executor import register
    from mars.remote.core import RemoteFunction
    from .core import execute_with_odps_context
    from .run_script import RunScript

    register(RemoteFunction, execute_with_odps_context(RemoteFunction.execute))
    register(RunScript, execute_with_odps_context(RunScript.execute))
except ImportError:
    pass


# register filesystem
try:
    from mars.filesystem import file_systems
    from .filesystem import VolumeFileSystem

    file_systems["odps"] = VolumeFileSystem
except ImportError:
    pass

try:
    from mars.lib.filesystem.core import register_filesystem
    from .filesystem import VolumeFileSystem

    register_filesystem("odps", VolumeFileSystem)
except ImportError:
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import sys
import threading

from ..compat import izip, six, futures
from ..utils import with_wait_argument
from .backends.core import Engine

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable


class Delay(object):
    def __init__(self):
        self._idx = 0
        self._lock = threading.RLock()
        self._calls = dict()
        self._futures = dict()
        self._wrappers = dict()
        self._running = False

    @with_wait_argument
    def execute(self, ui=None, async_=False, n_parallel=1, timeout=None,
                close_and_notify=True):
        if self._running:
            raise RuntimeError('Cannot execute on an executing delay object.')

        with self._lock:
            _idx_calls = sorted(six.iteritems(self._calls), key=lambda p: p[0])
            _indices = [p[0] for p in _idx_calls]
            _calls = [p[1] for p in _idx_calls]
            _futures = [p[1] for p in sorted(six.iteritems(self._futures), key=lambda p: p[0])]

        from .backends.engine import get_default_engine
        engine = get_default_engine(*[call[1] for call in _calls])

        ui = ui or Engine._create_ui(async_=async_, n_parallel=n_parallel)

        batch_kw = dict(ui=ui, async_=True, n_parallel=n_parallel, timeout=timeout,
                        close_and_notify=close_and_notify)
        fs = engine.batch(*_calls, **batch_kw)

        if not isinstance(fs, Iterable):
            fs = [fs]
        if len(fs) == 0:
            return

        self._running = True

        def relay_future(src, index=None, dest=None):
            if dest.done():
                return
            with self._lock:
                try:
                    result = src.result()
                    wrapper = self._wrappers.get(index)
                    if wrapper is not None:
                        result = wrapper(result)
                    dest.set_result(result)
                except:
                    e, tb = sys.exc_info()[1:]
                    if six.PY2:
                        dest.set_exception_info(e, tb)
                    else:
                        dest.set_exception(e)
                finally:
                    del self._calls[index]
                    del self._futures[index]
                    if index in self._wrappers:
                        del self._wrappers[index]

        for idx, uf, bf in izip(_indices, _futures, fs):
            uf.set_running_or_notify_cancel()
            bf.add_done_callback(functools.partial(relay_future, index=idx, dest=uf))
            if bf.done():
                relay_future(bf, idx, uf)

        if not async_:
            try:
                futures.wait(_futures, timeout)
                for uf in _futures:
                    uf.result()
                ui.notify('DataFrame execution succeeded')
            except:
                ui.notify('DataFrame execution failed')
                raise
            finally:
                self._running = False
        else:
            delay_future = futures.Future()
            delay_future.set_running_or_notify_cancel()
            result_store, exc_store = [None], [None]

            def _check_results(src):
                if delay_future.done():
                    return
                try:
                    result_store[0] = src.result()
                except:
                    exc_store[0] = sys.exc_info()[1:]

                with self._lock:
                    if all(f.done() for f in fs):
                        self._running = False
                        if exc_store[0] is not None:
                            e, tb = exc_store[0]
                            if six.PY2:
                                delay_future.set_exception_info(e, tb)
                            else:
                                delay_future.set_exception(e)
                        else:
                            delay_future.set_result(result_store[0])
            for f in fs:
                f.add_done_callback(_check_results)
                if f.done():
                    _check_results(f)

            return delay_future

    def register_item(self, action, expr, *args, **kwargs):
        wrapper = kwargs.pop('wrapper', None)
        with self._lock:
            call_idx = self._idx
            self._idx += 1
            user_future = futures.Future()
            self._calls[call_idx] = (action, expr, args, kwargs)
            self._futures[call_idx] = user_future
            if wrapper:
                self._wrappers[call_idx] = wrapper

        return user_future



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import serializers
from ..core import LazyLoad


class SecurityConfiguration(LazyLoad):
    _root = 'SecurityConfiguration'

    class ProjectProtection(serializers.XMLSerializableModel):
        _root = 'ProjectProtection'
        protected = serializers.XMLNodeAttributeField(attr='Protected', type='bool')
        exception = serializers.XMLNodeField('Exceptions')

    check_permission_using_acl = serializers.XMLNodeField('CheckPermissionUsingAcl', type='bool')
    check_permission_using_policy = serializers.XMLNodeField('CheckPermissionUsingPolicy', type='bool')
    label_security = serializers.XMLNodeField('LabelSecurity', type='bool')
    object_creator_has_access_permission = serializers.XMLNodeField('ObjectCreatorHasAccessPermission', type='bool')
    object_creator_has_grant_permission = serializers.XMLNodeField('ObjectCreatorHasGrantPermission', type='bool')
    project_protection = serializers.XMLNodeReferenceField(ProjectProtection, 'ProjectProtection')
    check_permission_using_acl_v2 = serializers.XMLNodeField('CheckPermissionUsingAclV2', type='bool')
    check_permission_using_policy_v2 = serializers.XMLNodeField('CheckPermissionUsingPolicyV2', type='bool')
    support_acl = serializers.XMLNodeField('SupportACL', type='bool')
    support_policy = serializers.XMLNodeField('SupportPolicy', type='bool')
    support_package = serializers.XMLNodeField('SupportPackage', type='bool')
    support_acl_v2 = serializers.XMLNodeField('SupportACLV2', type='bool')
    support_package_v2 = serializers.XMLNodeField('SupportPackageV2', type='bool')
    check_permission_using_package = serializers.XMLNodeField('CheckPermissionUsingPackage', type='bool')
    create_package = serializers.XMLNodeField('CreatePackage', type='bool')
    create_package_v2 = serializers.XMLNodeField('CreatePackageV2', type='bool')

    def reload(self):
        resp = self._client.get(
            self.project.resource(), params=dict(security_configuration='')
        )
        self.parse(self._client, resp, obj=self)

    def update(self):
        content = self.serialize()
        self._client.put(
            self.project.resource(), params=dict(security_configuration=''), data=content
        )



#!/usr/bin/env python

import os
import sys

from odps.udf.tools import runners


def _chr_if_necessary(s):
    try:
        n = int(s)
        return chr(n)
    except ValueError:
        return s


def main():
    sys.path.insert(0, os.getcwd())

    from odps import udf
    # Arguments parsing
    import argparse
    parser = argparse.ArgumentParser(description='ODPS Python UDF tools')
    parser.add_argument('-D', metavar='delim', type=str, default=',',
                        help='Line delimiter that separates lines into columns, '
                        'default is ","')
    parser.add_argument('-N', metavar='null', type=str, default='NULL',
                        help='NULL indicator')
    parser.add_argument('-I', metavar='stdin', type=str, default='sys.stdin',
                        help='standard input, sys.stdin as default')
    parser.add_argument('clz', metavar='your_script.class_name', type=str, help='The full import path of your UDF class')
    args = parser.parse_args()
    delim = _chr_if_necessary(args.D)
    null_indicator = _chr_if_necessary(args.N)

    # Import user class
    pkg, name = args.clz.rsplit('.', 1)
    usermod = __import__(pkg, globals(), locals(), [name])
    clz = getattr(usermod, name)

    # get stdin
    pkg, name = args.I.rsplit('.', 1)
    usermod = __import__(pkg, globals(), locals(), [name])
    stdin = getattr(usermod, name)

    udf_runner = runners.get_default_runner(clz, delim, null_indicator, stdin)
    udf_runner.run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import os
import sys
from hashlib import md5
import pickle
import shutil

from .compat import six
from .config import options
from .core import ODPS
from .errors import InteractiveError
from .models import TableSchema
from .utils import to_binary, build_pyodps_dir
from .df.backends.frame import ResultFrame
from .df.backends.odpssql.types import odps_schema_to_df_schema


DEFAULT_ROOM_NAME = 'default'
ODPS_FILE_NAME = '__ODPS__'
INFO_FILE_NAME = '__INFO__'
OBJECT_FILE_NAME = '__OBJ__'


class Room(object):
    def __init__(self, room_name):
        self._room_name = room_name
        self._room_dir = _get_room_dir(self._room_name)

        self._inited = False
        self._init()

    def _init(self):
        if self._inited:
            return

        odps_file = os.path.join(self._room_dir, ODPS_FILE_NAME)
        if not os.path.exists(odps_file):
            raise InteractiveError(
                'This room(%s) is not configured' % self._room_name)

        with open(odps_file, 'rb') as f:
            try:
                obj = pickle.load(f)
            except pickle.UnpicklingError:
                raise InteractiveError(
                    'Failed to enter a room: %s' % self._room_name)

            def _config_rooms(access_id, access_key, default_project, endpoint, tunnel_endpoint=None,
                              seahawks_url=None, logview_host=None, default_schema=None,
                              region_name=None, **kwargs):
                options.loads(kwargs.get('options', {}))

                options.account = ODPS._build_account(access_id, access_key)
                options.endpoint = endpoint
                options.default_project = default_project
                options.default_schema = default_schema
                options.region_name = region_name
                options.tunnel.endpoint = tunnel_endpoint
                options.logview_host = logview_host
                options.seahawks_url = seahawks_url

            if isinstance(obj[-1], dict):
                kw = obj[-1]
                obj = obj[:-1]
            else:
                kw = dict()
            _config_rooms(*obj, **kw)

        self._inited = True

    @property
    def odps(self):
        return ODPS._from_account(
            options.account,
            options.default_project,
            schema=options.default_schema,
            endpoint=options.endpoint,
            region_name=options.region_name,
            tunnel_endpoint=options.tunnel.endpoint,
            logview_host=options.logview_host,
            seahawks_url=options.seahawks_url,
        )

    def __getattr__(self, attr):
        try:
            return self.fetch(attr)
        except InteractiveError:
            return super(Room, self).__getattr__(attr)

    def __getitem__(self, item):
        return self.fetch(item)

    def __delitem__(self, key):
        self.drop(key)

    def _obj_store_dir(self, name):
        fn = md5(to_binary(name)).hexdigest()
        return os.path.join(self._room_dir, fn)

    def store(self, name, obj, desc=None):
        path = self._obj_store_dir(name)

        if os.path.exists(path):
            raise InteractiveError('%s already exists' % name)

        os.makedirs(path)
        with open(os.path.join(path, INFO_FILE_NAME), 'wb') as f:
            pickle.dump((name, desc), f, protocol=0)

        with open(os.path.join(path, OBJECT_FILE_NAME), 'wb') as f:
            pickle.dump(obj, f, protocol=0)

    def fetch(self, name):
        path = self._obj_store_dir(name)

        if not os.path.exists(path):
            raise InteractiveError('%s does not exist' % name)

        with open(os.path.join(path, OBJECT_FILE_NAME), 'rb') as f:
            return pickle.load(f)

    def drop(self, name):
        path = self._obj_store_dir(name)

        if os.path.exists(path):
            shutil.rmtree(path)

    def list_stores(self):
        results = []
        for obj_dir in os.listdir(self._room_dir):
            info_path = os.path.join(self._room_dir, obj_dir, INFO_FILE_NAME)
            if os.path.exists(info_path):
                with open(info_path, 'rb') as f:
                    results.append(list(pickle.load(f)))

        return results

    def display(self):
        schema = TableSchema.from_lists(['name', 'desc'], ['string'] * 2)
        schema = odps_schema_to_df_schema(schema)
        frame = ResultFrame(self.list_stores(), schema=schema, pandas=False)
        try:
            import pandas as pd

            if isinstance(frame.values, pd.DataFrame):
                df = frame.values

                df.columns.name = self._room_name
                frame._values = df.set_index('name')
        except ImportError:
            pass

        return frame


def _get_root_dir():
    rooms_dir = build_pyodps_dir('rooms')

    return os.path.join(rooms_dir, str(sys.version_info[0]))


def _get_room_dir(room_name, mkdir=False):
    rooms_dir = _get_root_dir()

    room_name = md5(to_binary(room_name)).hexdigest()
    room_dir = os.path.join(rooms_dir, room_name)
    if not os.path.exists(room_dir) and mkdir:
        os.makedirs(room_dir)

    return room_dir


def setup(access_id, access_key, default_project, endpoint=None, tunnel_endpoint=None,
          default_schema=None, region_name=None, seahawks_url=None, logview_host=None,
          room=DEFAULT_ROOM_NAME, with_options=False, **kwargs):
    room_dir = _get_room_dir(room, mkdir=True)
    odps_file = os.path.join(room_dir, ODPS_FILE_NAME)

    if with_options:
        trivial_types = (six.string_types, six.integer_types, float, type(None))
        options_dump = {
            k: v for k, v in six.iteritems(options.dumps()) if isinstance(v, trivial_types)
        }
        kwargs['options'] = options_dump

    if os.path.exists(odps_file):
        raise InteractiveError(
            'This room(%s) has been configured before, '
            'you can teardown it first' % room)

    obj = (access_id, access_key, default_project, endpoint, tunnel_endpoint,
           seahawks_url, logview_host, default_schema, region_name, kwargs)

    with open(odps_file, 'wb') as f:
        pickle.dump(obj, f, protocol=0)

    with open(os.path.join(room_dir, INFO_FILE_NAME), 'wb') as f:
        f.write(to_binary(room))


def enter(room=DEFAULT_ROOM_NAME):
    return Room(room)


def teardown(room=DEFAULT_ROOM_NAME):
    room_dir = _get_room_dir(room)
    if os.path.exists(room_dir):
        shutil.rmtree(room_dir)


def list_rooms():
    root = _get_root_dir()

    results = []
    for room_dir in os.listdir(root):
        info_path = os.path.join(root, room_dir, INFO_FILE_NAME)
        if os.path.exists(info_path):
            with open(info_path) as f:
                results.append(f.read())

    return results
